"""unidep - Unified Conda and Pip requirements management.

This module contains the Hatchling integration.
"""

from __future__ import annotations

from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.plugin import hookimpl

from unidep._setuptools_integration import _deps
from unidep.utils import (
    parse_folder_or_filename,
)

__all__ = ["UnidepRequirementsMetadataHook"]


class UnidepRequirementsMetadataHook(MetadataHookInterface):
    """Hatch hook to populate ``'project.depencencies'`` from ``requirements.yaml`` or ``pyproject.toml``."""  # noqa: E501

    PLUGIN_NAME = "unidep"

    def update(self, metadata: dict) -> None:
        """Update the project table's metadata."""
        if "dependencies" not in metadata.get("dynamic", []):
            return
        project_root = Path.cwd()
        try:
            requirements_file = parse_folder_or_filename(project_root).path
        except FileNotFoundError:
            return
        if "dependencies" in metadata:
            error_msg = (
                "You have a `requirements.yaml` file in your project root or"
                " configured unidep in `pyproject.toml` with `[tool.unidep]`,"
                " but you are also using `[project.dependencies]`."
                " Please remove `[project.dependencies]`, you cannot use both."
            )
            raise RuntimeError(error_msg)

        deps = _deps(requirements_file)
        metadata["dependencies"] = deps.dependencies
        if "optional-dependencies" not in metadata.get("dynamic", []):
            return
        metadata["optional-dependencies"] = deps.extras


@hookimpl
def hatch_register_metadata_hook() -> type[UnidepRequirementsMetadataHook]:
    return UnidepRequirementsMetadataHook



"""unidep - Unified Conda and Pip requirements management.

This module provides utility functions used throughout the package.
"""

from __future__ import annotations

import codecs
import platform
import re
import sys
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, NamedTuple, cast

from unidep._version import __version__
from unidep.platform_definitions import (
    PEP508_MARKERS,
    Platform,
    Selector,
    platforms_from_selector,
    validate_selector,
)

if sys.version_info >= (3, 11):
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib


def add_comment_to_file(
    filename: str | Path,
    extra_lines: list[str] | None = None,
) -> None:
    """Add a comment to the top of a file."""
    if extra_lines is None:
        extra_lines = []
    with open(filename, "r+") as f:  # noqa: PTH123
        content = f.read()
        f.seek(0, 0)
        command_line_args = " ".join(sys.argv[1:])
        txt = [
            f"# This file is created and managed by `unidep` {__version__}.",
            "# For details see https://github.com/basnijholt/unidep",
            f"# File generated with: `unidep {command_line_args}`",
            *extra_lines,
        ]
        content = "\n".join(txt) + "\n\n" + content
        f.write(content)


def remove_top_comments(filename: str | Path) -> None:
    """Removes the top comments (lines starting with '#') from a file."""
    with open(filename) as file:  # noqa: PTH123
        lines = file.readlines()

    first_non_comment = next(
        (i for i, line in enumerate(lines) if not line.strip().startswith("#")),
        len(lines),
    )
    content_without_comments = lines[first_non_comment:]
    with open(filename, "w") as file:  # noqa: PTH123
        file.writelines(content_without_comments)


def escape_unicode(string: str) -> str:
    """Escape unicode characters."""
    return codecs.decode(string, "unicode_escape")


def is_pip_installable(folder: str | Path) -> bool:  # pragma: no cover
    """Determine if the project is pip installable.

    Checks for existence of setup.py or [build-system] in pyproject.toml.
    If the `toml` library is available, it is used to parse the `pyproject.toml` file.
    If the `toml` library is not available, the function checks for the existence of
    a line starting with "[build-system]". This does not handle the case where
    [build-system] is inside of a multi-line literal string.
    """
    path = Path(folder)
    if (path / "setup.py").exists():
        return True

    pyproject_path = path / "pyproject.toml"
    if pyproject_path.exists():
        with pyproject_path.open("rb") as file:
            pyproject_data = tomllib.load(file)
            return "build-system" in pyproject_data
    return False


class UnsupportedPlatformError(Exception):
    """Raised when the current platform is not supported."""


def identify_current_platform() -> Platform:
    """Detect the current platform."""
    system = platform.system().lower()
    architecture = platform.machine().lower()

    if system == "linux":
        if architecture == "x86_64":
            return "linux-64"
        if architecture == "aarch64":
            return "linux-aarch64"
        if architecture == "ppc64le":
            return "linux-ppc64le"
        msg = f"Unsupported Linux architecture `{architecture}`"
        raise UnsupportedPlatformError(msg)
    if system == "darwin":
        if architecture == "x86_64":
            return "osx-64"
        if architecture == "arm64":
            return "osx-arm64"
        msg = f"Unsupported macOS architecture `{architecture}`"
        raise UnsupportedPlatformError(msg)
    if system == "windows":
        if "64" in architecture:
            return "win-64"
        msg = f"Unsupported Windows architecture `{architecture}`"
        raise UnsupportedPlatformError(msg)
    msg = f"Unsupported operating system `{system}` with architecture `{architecture}`"
    raise UnsupportedPlatformError(msg)


def build_pep508_environment_marker(
    platforms: list[Platform | tuple[Platform, ...]],
) -> str:
    """Generate a PEP 508 selector for a list of platforms."""
    sorted_platforms = tuple(sorted(platforms))
    if sorted_platforms in PEP508_MARKERS:
        return PEP508_MARKERS[sorted_platforms]  # type: ignore[index]
    environment_markers = [
        PEP508_MARKERS[platform]
        for platform in sorted(sorted_platforms)
        if platform in PEP508_MARKERS
    ]
    return " or ".join(environment_markers)


class ParsedPackageStr(NamedTuple):
    """A package name and version pinning."""

    name: str
    pin: str | None = None
    # can be of type `Selector` but also space separated string of `Selector`s
    selector: str | None = None


def parse_package_str(package_str: str) -> ParsedPackageStr:
    """Splits a string into package name, version pinning, and platform selector."""
    # Regex to match package name, version pinning, and optionally platform selector
    # Note: the name_pattern currently allows for paths and extras, however,
    # paths cannot contain spaces or contain brackets.
    name_pattern = r"[a-zA-Z0-9_.\-/]+(\[[a-zA-Z0-9_.,\-]+\])?"
    version_pin_pattern = r".*?"
    selector_pattern = r"[a-z0-9\s]+"
    pattern = rf"({name_pattern})\s*({version_pin_pattern})?(:({selector_pattern}))?$"
    match = re.match(pattern, package_str)

    if match:
        package_name = match.group(1).strip()
        version_pin = match.group(3).strip() if match.group(3) else None
        selector = match.group(5).strip() if match.group(5) else None

        if selector is not None:
            for s in selector.split():
                validate_selector(cast(Selector, s))

        return ParsedPackageStr(
            package_name,
            version_pin,
            selector,
        )

    msg = f"Invalid package string: '{package_str}'"
    raise ValueError(msg)


def _simple_warning_format(
    message: Warning | str,
    category: type[Warning],  # noqa: ARG001
    filename: str,
    lineno: int,
    line: str | None = None,  # noqa: ARG001
) -> str:  # pragma: no cover
    """Format warnings without code context."""
    return (
        f"---------------------\n"
        f"⚠️  *** WARNING *** ⚠️\n"
        f"{message}\n"
        f"Location: {filename}:{lineno}\n"
        f"---------------------\n"
    )


def warn(
    message: str | Warning,
    category: type[Warning] = UserWarning,
    stacklevel: int = 1,
) -> None:
    """Emit a warning with a custom format specific to this package."""
    original_format = warnings.formatwarning
    warnings.formatwarning = _simple_warning_format
    try:
        warnings.warn(message, category, stacklevel=stacklevel + 1)
    finally:
        warnings.formatwarning = original_format


def selector_from_comment(comment: str) -> str | None:
    """Extract a valid selector from a comment."""
    multiple_brackets_pat = re.compile(r"#.*\].*\[")  # Detects multiple brackets
    if multiple_brackets_pat.search(comment):
        msg = f"Multiple bracketed selectors found in comment: '{comment}'"
        raise ValueError(msg)

    sel_pat = re.compile(r"#\s*\[([^\[\]]+)\]")
    m = sel_pat.search(comment)
    if not m:
        return None
    selectors = m.group(1).strip().split()
    for s in selectors:
        validate_selector(cast(Selector, s))
    return " ".join(selectors)


def extract_matching_platforms(comment: str) -> list[Platform]:
    """Get all platforms matching a comment."""
    selector = selector_from_comment(comment)
    if selector is None:
        return []
    return platforms_from_selector(selector)


def unidep_configured_in_toml(path: Path) -> bool:
    """Check if dependencies are specified in pyproject.toml."""
    with path.open("rb") as f:
        data = tomllib.load(f)
    return bool(data.get("tool", {}).get("unidep", {}))


def split_path_and_extras(input_str: str | Path) -> tuple[Path, list[str]]:
    """Parse a string of the form `path/to/file[extra1,extra2]` into parts.

    Returns a tuple of the `pathlib.Path` and a list of extras
    """
    if isinstance(input_str, Path):
        input_str = str(input_str)

    if not input_str:  # Check for empty string
        return Path(), []

    pattern = r"^(.+?)(?:\[([^\[\]]+)\])?$"
    match = re.search(pattern, input_str)

    if match is None:  # pragma: no cover
        # I don't think this is possible, but just in case
        return Path(), []

    path = Path(match.group(1))
    extras = match.group(2)
    if not extras:
        return path, []
    extras = [extra.strip() for extra in extras.split(",")]
    return path, extras


class PathWithExtras(NamedTuple):
    """A dependency file and extras."""

    path: Path
    extras: list[str]

    @property
    def path_with_extras(self) -> Path:
        """Path including extras, e.g., `path/to/file[test,docs]`."""
        if not self.extras:
            return self.path
        return Path(f"{self.path}[{','.join(self.extras)}]")

    def resolved(self) -> PathWithExtras:
        """Resolve the path and extras."""
        return PathWithExtras(self.path.resolve(), self.extras)

    def __hash__(self) -> int:
        """Hash the path and extras."""
        return hash((self.path, tuple(sorted(self.extras))))

    def __eq__(self, other: object) -> bool:
        """Check if two `PathWithExtras` are equal."""
        if not isinstance(other, PathWithExtras):
            return NotImplemented
        return self.path == other.path and set(self.extras) == set(other.extras)


def parse_folder_or_filename(folder_or_file: str | Path) -> PathWithExtras:
    """Get the path to `requirements.yaml` or `pyproject.toml` file."""
    folder_or_file, extras = split_path_and_extras(folder_or_file)
    path = Path(folder_or_file)
    if path.is_dir():
        fname_yaml = path / "requirements.yaml"
        if fname_yaml.exists():
            return PathWithExtras(fname_yaml, extras)
        fname_toml = path / "pyproject.toml"
        if fname_toml.exists() and unidep_configured_in_toml(fname_toml):
            return PathWithExtras(fname_toml, extras)
        msg = (
            f"File `{fname_yaml}` or `{fname_toml}` (with unidep configuration)"
            f" not found in `{folder_or_file}`."
        )
        raise FileNotFoundError(msg)
    if not path.exists():
        msg = f"File `{path}` not found."
        raise FileNotFoundError(msg)
    return PathWithExtras(path, extras)


def defaultdict_to_dict(d: defaultdict | Any) -> dict:
    """Convert (nested) defaultdict to (nested) dict."""
    if isinstance(d, defaultdict):
        d = {key: defaultdict_to_dict(value) for key, value in d.items()}
    return d


def get_package_version(package_name: str) -> str | None:
    """Returns the version of the given package.

    Parameters
    ----------
    package_name
        The name of the package to find the version of.

    Returns
    -------
    The version of the package, or None if the package is not found.

    """
    if sys.version_info >= (3, 8):
        import importlib.metadata

        try:
            return importlib.metadata.version(package_name)
        except importlib.metadata.PackageNotFoundError:
            return None
    else:  # pragma: no cover
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            import pkg_resources

        try:
            return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            return None



#!/usr/bin/env python3
"""unidep - Unified Conda and Pip requirements management.

This module provides a command-line tool for managing conda environment.yaml files.
"""

from __future__ import annotations

import argparse
import functools
import importlib.util
import itertools
import json
import os
import platform
import shutil
import subprocess
import sys
import time
from pathlib import Path

from unidep._conda_env import (
    create_conda_env_specification,
    write_conda_environment_file,
)
from unidep._conda_lock import conda_lock_command
from unidep._conflicts import resolve_conflicts
from unidep._dependencies_parsing import (
    find_requirements_files,
    parse_local_dependencies,
    parse_requirements,
)
from unidep._setuptools_integration import (
    filter_python_dependencies,
    get_python_dependencies,
)
from unidep._version import __version__
from unidep.platform_definitions import Platform
from unidep.utils import (
    add_comment_to_file,
    escape_unicode,
    get_package_version,
    identify_current_platform,
    is_pip_installable,
    parse_folder_or_filename,
    parse_package_str,
    warn,
)

if sys.version_info >= (3, 8):
    from typing import Literal, get_args
else:  # pragma: no cover
    from typing_extensions import Literal, get_args

try:  # pragma: no cover
    from rich_argparse import RichHelpFormatter

    class _HelpFormatter(RichHelpFormatter):
        def _get_help_string(self, action: argparse.Action) -> str | None:
            # escapes "[" in text, otherwise e.g., [linux] is removed
            if action.help is not None:
                return action.help.replace("[", r"\[")
            return None
except ImportError:  # pragma: no cover
    from argparse import HelpFormatter as _HelpFormatter  # type: ignore[assignment]

_DEP_FILES = "`requirements.yaml` or `pyproject.toml`"
CondaExecutable = Literal["conda", "mamba", "micromamba"]


def _add_common_args(  # noqa: PLR0912, C901
    sub_parser: argparse.ArgumentParser,
    options: set[str],
) -> None:  # pragma: no cover
    if "directory" in options:
        sub_parser.add_argument(
            "-d",
            "--directory",
            type=Path,
            default=".",
            help=f"Base directory to scan for {_DEP_FILES} file(s), by default `.`",
        )
    if "depth" in options:
        sub_parser.add_argument(
            "--depth",
            type=int,
            default=1,
            help=f"Maximum depth to scan for {_DEP_FILES} files, by default 1",
        )
    if "file" in options or "file-alt" in options:
        if "file-alt" in options:
            help_msg = (
                f"A single {_DEP_FILES} file to use, or"
                " folder that contains that file. This is an alternative to using"
                f" `--directory` which searches for all {_DEP_FILES} files in the"
                " directory and its subdirectories."
            )
        else:
            help_msg = (
                f"The {_DEP_FILES} file to parse, or folder"
                " that contains that file, by default `.`"
            )
        assert "conda-lock-file" not in options  # both use "-f"
        sub_parser.add_argument(
            "-f",
            "--file",
            type=Path,
            default=[],
            action="append",
            help=help_msg,
        )
    if "*files" in options:
        sub_parser.add_argument(
            "files",
            type=Path,
            nargs="+",
            help=f"The {_DEP_FILES} file(s) to parse"
            " or folder(s) that contain those file(s), by default `.`",
            default=None,
        )
    if "verbose" in options:
        sub_parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="Print verbose output",
        )
    if "platform" in options:
        current_platform = identify_current_platform()
        sub_parser.add_argument(
            "-p",
            "--platform",
            type=str,
            action="append",  # Allow multiple instances of -p
            default=[],
            choices=get_args(Platform),
            help="The platform(s) to get the requirements for. "
            "Multiple platforms can be specified. "
            f"By default, the current platform (`{current_platform}`) is used.",
        )
    if "editable" in options:
        sub_parser.add_argument(
            "-e",
            "--editable",
            action="store_true",
            help="Install the project in editable mode",
        )
    if "skip-local" in options:
        sub_parser.add_argument(
            "--skip-local",
            action="store_true",
            help="Skip installing local dependencies",
        )
    if "skip-pip" in options:
        sub_parser.add_argument(
            "--skip-pip",
            action="store_true",
            help=f"Skip installing pip dependencies from {_DEP_FILES}",
        )
    if "skip-conda" in options:
        sub_parser.add_argument(
            "--skip-conda",
            action="store_true",
            help=f"Skip installing conda dependencies from {_DEP_FILES}",
        )
    if "skip-dependency" in options:
        sub_parser.add_argument(
            "--skip-dependency",
            type=str,
            action="append",
            default=[],
            help="Skip installing a specific dependency that is in one of the"
            f" {_DEP_FILES}"
            " files. This option can be used multiple times, each"
            " time specifying a different package to skip."
            " For example, use `--skip-dependency pandas` to skip installing pandas.",
        )
    if "no-dependencies" in options:
        sub_parser.add_argument(
            "--no-dependencies",
            "--no-deps",
            action="store_true",
            help=f"Skip installing dependencies from {_DEP_FILES}"
            " file(s) and only install local package(s). Useful after"
            " installing a `conda-lock.yml` file because then all"
            " dependencies have already been installed.",
        )
    if "conda-executable" in options:
        sub_parser.add_argument(
            "--conda-executable",
            type=str,
            choices=("conda", "mamba", "micromamba"),
            help="The conda executable to use",
            default=None,
        )
    if "conda-env" in options:
        grp = sub_parser.add_mutually_exclusive_group()
        grp.add_argument(
            "-n",
            "--conda-env-name",
            type=str,
            default=None,
            help="Name of the conda environment, if not provided, the currently"
            " active environment name is used, unless `--conda-env-prefix` is"
            " provided",
        )
        grp.add_argument(
            "-p",  # Overlaps with `--platform`, but that's fine
            "--conda-env-prefix",
            type=Path,
            default=None,
            help="Path to the conda environment, if not provided, the currently"
            " active environment path is used, unless `--conda-env-name` is"
            " provided",
        )
    if "dry-run" in options:
        sub_parser.add_argument(
            "--dry-run",
            "--dry",
            action="store_true",
            help="Only print the commands that would be run",
        )
    if "ignore-pin" in options:
        sub_parser.add_argument(
            "--ignore-pin",
            type=str,
            action="append",
            default=[],
            help="Ignore the version pin for a specific package,"
            " e.g., `--ignore-pin numpy`. This option can be repeated"
            " to ignore multiple packages.",
        )
    if "overwrite-pin" in options:
        sub_parser.add_argument(
            "--overwrite-pin",
            type=str,
            action="append",
            default=[],
            help="Overwrite the version pin for a specific package,"
            " e.g., `--overwrite-pin 'numpy=1.19.2'`. This option can be repeated"
            " to overwrite the pins of multiple packages.",
        )
    if "conda-lock-file" in options:
        sub_parser.add_argument(
            "-f",
            "--conda-lock-file",
            type=Path,
            help="Path to the `conda-lock.yml` file to use for creating the new"
            " environment. Assumes that the lock file contains all dependencies."
            " Must be used with `--conda-env-name` or `--conda-env-prefix`.",
        )
    if "no-uv" in options:
        sub_parser.add_argument(
            "--no-uv",
            action="store_true",
            help="Disables the use of `uv` for pip install. By default, `uv` is used"
            " if it is available in the PATH.",
        )


def _add_extra_flags(
    subparser: argparse.ArgumentParser,
    downstream_command: str,
    unidep_subcommand: str,
    example: str,
) -> None:
    subparser.add_argument(
        "extra_flags",
        nargs=argparse.REMAINDER,
        help=f"Extra flags to pass to `{downstream_command}`. These flags are passed"
        f" directly and should be provided in the format expected by"
        f" `{downstream_command}`. For example, `unidep {unidep_subcommand} -- {example}`."  # noqa: E501
        f" Note that the `--` is required to separate the flags for"
        f" `unidep {unidep_subcommand}` from the flags for `{downstream_command}`.",
    )


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Unified Conda and Pip requirements management.",
        formatter_class=_HelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="command", help="Subcommands")

    # Subparser for the 'merge' command
    merge_help = (
        f"Combine multiple (or a single) {_DEP_FILES}"
        " files into a"
        " single Conda installable `environment.yaml` file."
    )
    merge_example = (
        " Example usage: `unidep merge --directory . --depth 1 --output environment.yaml`"  # noqa: E501
        f" to search for {_DEP_FILES}"
        " files in the current directory and its"
        " subdirectories and create `environment.yaml`. These are the defaults, so you"
        " can also just run `unidep merge`."
    )
    parser_merge = subparsers.add_parser(
        "merge",
        help=merge_help,
        description=merge_help + merge_example,
        formatter_class=_HelpFormatter,
    )
    parser_merge.add_argument(
        "-o",
        "--output",
        type=Path,
        default="environment.yaml",
        help="Output file for the conda environment, by default `environment.yaml`",
    )
    parser_merge.add_argument(
        "-n",
        "--name",
        type=str,
        default="myenv",
        help="Name of the conda environment, by default `myenv`",
    )
    parser_merge.add_argument(
        "--stdout",
        action="store_true",
        help="Output to stdout instead of a file",
    )
    parser_merge.add_argument(
        "--selector",
        type=str,
        choices=("sel", "comment"),
        default="sel",
        help="The selector to use for the environment markers, if `sel` then"
        " `- numpy # [linux]` becomes `sel(linux): numpy`, if `comment` then"
        " it remains `- numpy # [linux]`, by default `sel`",
    )
    _add_common_args(
        parser_merge,
        {
            "directory",
            "verbose",
            "platform",
            "depth",
            "ignore-pin",
            "skip-dependency",
            "overwrite-pin",
        },
    )

    # Subparser for the 'install' command
    install_help = (
        f"Automatically install all dependencies from one or more {_DEP_FILES} files."
        " This command first installs dependencies"
        " with Conda, then with Pip. Finally, it installs local packages"
        f" (those containing the {_DEP_FILES} files)"
        " using `pip install [-e] ./project`."
    )
    install_example = (
        " Example usage: `unidep install .` for a single project."
        " For multiple projects: `unidep install ./project1 ./project2`."
        " The command accepts both file paths and directories containing"
        f" a {_DEP_FILES} file. Use `--editable` or"
        " `-e` to install the local packages in editable mode. See"
        f" `unidep install-all` to install all {_DEP_FILES} files in and below the"
        " current folder."
    )

    parser_install = subparsers.add_parser(
        "install",
        help=install_help,
        description=install_help + install_example,
        formatter_class=_HelpFormatter,
    )

    # Add positional argument for the file
    _add_common_args(
        parser_install,
        {
            "*files",
            "conda-executable",
            "conda-env",
            "conda-lock-file",
            "dry-run",
            "editable",
            "skip-local",
            "skip-pip",
            "skip-conda",
            "no-dependencies",
            "ignore-pin",
            "skip-dependency",
            "overwrite-pin",
            "no-uv",
            "verbose",
        },
    )
    install_all_help = (
        f"Install dependencies from all {_DEP_FILES}"
        " files found in the current"
        " directory or specified directory. This command first installs dependencies"
        " using Conda, then Pip, and finally the local packages."
    )
    install_all_example = (
        " Example usage: `unidep install-all` to install dependencies from all"
        f" {_DEP_FILES}"
        " files in the current directory. Use"
        " `--directory ./path/to/dir` to specify a different directory. Use"
        " `--depth` to control the depth of directory search. Add `--editable`"
        " or `-e` for installing local packages in editable mode."
    )

    parser_install_all = subparsers.add_parser(
        "install-all",
        help=install_all_help,
        description=install_all_help + install_all_example,
        formatter_class=_HelpFormatter,
    )

    # Add positional argument for the file
    _add_common_args(
        parser_install_all,
        {
            "conda-executable",
            "conda-env",
            "conda-lock-file",
            "dry-run",
            "editable",
            "depth",
            "directory",
            "skip-local",
            "skip-pip",
            "skip-conda",
            "no-dependencies",
            "ignore-pin",
            "skip-dependency",
            "overwrite-pin",
            "no-uv",
            "verbose",
        },
    )

    # Subparser for the 'conda-lock' command

    conda_lock_help = (
        "Generate a global `conda-lock.yml` file for a collection of"
        f" {_DEP_FILES}"
        " files. Additionally, create individual"
        f" `conda-lock.yml` files for each {_DEP_FILES} file"
        " consistent with the global lock file."
    )
    conda_lock_example = (
        " Example usage: `unidep conda-lock --directory ./projects` to generate"
        f" conda-lock files for all {_DEP_FILES}"
        " files in the `./projects`"
        " directory. Use `--only-global` to generate only the global lock file."
        " The `--check-input-hash` option can be used to avoid regenerating lock"
        " files if the input hasn't changed."
    )

    parser_lock = subparsers.add_parser(
        "conda-lock",
        help=conda_lock_help,
        description=conda_lock_help + conda_lock_example,
        formatter_class=_HelpFormatter,
    )

    parser_lock.add_argument(
        "--only-global",
        action="store_true",
        help="Only generate the global lock file",
    )
    parser_lock.add_argument(
        "--lockfile",
        type=Path,
        default="conda-lock.yml",
        help="Specify a path for the global lockfile (default: `conda-lock.yml`"
        " in current directory). Path should be relative, e.g.,"
        " `--lockfile ./locks/example.conda-lock.yml`.",
    )
    parser_lock.add_argument(
        "--check-input-hash",
        action="store_true",
        help="Check existing input hashes in lockfiles before regenerating lock files."
        " This flag is directly passed to `conda-lock`.",
    )
    _add_common_args(
        parser_lock,
        {
            "directory",
            "file-alt",
            "verbose",
            "platform",
            "depth",
            "ignore-pin",
            "skip-dependency",
            "overwrite-pin",
        },
    )
    _add_extra_flags(parser_lock, "conda-lock lock", "conda-lock", "--micromamba")

    # Subparser for the 'pip-compile' command
    pip_compile_help = (
        "Generate a fully pinned `requirements.txt` file from one or more"
        f" {_DEP_FILES}"
        " files using `pip-compile` from `pip-tools`. This"
        f" command consolidates all pip dependencies defined in the {_DEP_FILES}"
        " files and compiles them into a single `requirements.txt` file, taking"
        " into account the specific versions and dependencies of each package."
    )
    pip_compile_example = (
        " Example usage: `unidep pip-compile --directory ./projects` to generate"
        f" a `requirements.txt` file for all {_DEP_FILES}"
        " files in the"
        " `./projects` directory. Use `--output-file requirements.txt` to specify a"
        " different output file."
    )

    parser_pip_compile = subparsers.add_parser(
        "pip-compile",
        help=pip_compile_help,
        description=pip_compile_help + pip_compile_example,
        formatter_class=_HelpFormatter,
    )
    parser_pip_compile.add_argument(
        "-o",
        "--output-file",
        type=Path,
        default=None,
        help="Output file for the pip requirements, by default `requirements.txt`",
    )
    _add_common_args(
        parser_pip_compile,
        {
            "directory",
            "verbose",
            "platform",
            "depth",
            "ignore-pin",
            "skip-dependency",
            "overwrite-pin",
        },
    )
    _add_extra_flags(
        parser_pip_compile,
        "pip-compile",
        "pip-compile",
        "--generate-hashes --allow-unsafe",
    )

    # Subparser for the 'pip' and 'conda' command
    help_str = "Get the {} requirements for the current platform only."
    help_example = (
        " Example usage: `unidep {which} --file folder1 --file"
        " folder2/requirements.yaml --separator ' ' --platform linux-64` to"
        " extract all the {which} dependencies specific to the linux-64 platform. Note"
        " that the `--file` argument can be used multiple times to specify multiple"
        f" {_DEP_FILES}"
        " files and that --file can also be a folder that contains"
        f" a {_DEP_FILES} file."
    )
    parser_pip = subparsers.add_parser(
        "pip",
        help=help_str.format("pip"),
        description=help_str.format("pip") + help_example.format(which="pip"),
        formatter_class=_HelpFormatter,
    )
    parser_conda = subparsers.add_parser(
        "conda",
        help=help_str.format("conda"),
        description=help_str.format("conda") + help_example.format(which="conda"),
        formatter_class=_HelpFormatter,
    )
    for sub_parser in [parser_pip, parser_conda]:
        _add_common_args(
            sub_parser,
            {
                "verbose",
                "platform",
                "file",
                "ignore-pin",
                "skip-dependency",
                "overwrite-pin",
            },
        )
        sub_parser.add_argument(
            "--separator",
            type=str,
            default=" ",
            help="The separator between the dependencies, by default ` `",
        )

    # Subparser for the 'version' command
    parser_merge = subparsers.add_parser(
        "version",
        help="Print version information of unidep.",
        formatter_class=_HelpFormatter,
    )

    args = parser.parse_args()

    if args.command is None:  # pragma: no cover
        parser.print_help()
        sys.exit(1)

    if "file" in args:
        _ensure_files(args.file)

    return args


def _ensure_files(files: list[Path]) -> None:
    """Ensure that the files exist."""
    missing = []
    for i, f in enumerate(files):
        try:
            path_with_extras = parse_folder_or_filename(f)
        except FileNotFoundError:  # noqa: PERF203
            missing.append(f"`{f}`")
        else:
            files[i] = path_with_extras.path_with_extras
    if missing:
        print(f"❌ One or more files ({', '.join(missing)}) not found.")
        sys.exit(1)


def _get_conda_executable(which: CondaExecutable) -> str | None:
    if shutil.which(which):
        return which  # Found in PATH so return the name
    # e.g., micromamba might be a bash function, check env var in that case
    env_var = "CONDA_EXE" if which == "conda" else "MAMBA_EXE"
    exe = os.environ.get(env_var, None)
    if exe is None:  # pragma: no cover
        return None
    if Path(exe).name != which:  # pragma: no cover
        return None
    return exe


def _identify_conda_executable() -> CondaExecutable:  # pragma: no cover
    """Identify the conda executable to use.

    This function checks for micromamba, mamba, and conda in that order.
    """
    if _get_conda_executable("micromamba") is not None:
        return "micromamba"
    if _get_conda_executable("mamba") is not None:
        return "mamba"
    if _get_conda_executable("conda") is not None:
        return "conda"
    msg = "Could not identify conda executable."
    raise RuntimeError(msg)


def _maybe_conda_executable() -> CondaExecutable | None:
    try:
        return _identify_conda_executable()
    except RuntimeError:  # pragma: no cover
        return None


def _format_inline_conda_package(package: str) -> str:
    pkg = parse_package_str(package)
    if pkg.pin is None:
        return pkg.name
    return f'{pkg.name}"{pkg.pin.strip()}"'


def _maybe_exe(conda_executable: CondaExecutable) -> str:
    """Add .exe on Windows."""
    if os.name == "nt":  # pragma: no cover
        if conda_executable in ("micromamba", "mamba") and os.environ.get("MAMBA_EXE"):
            return os.environ["MAMBA_EXE"]
        if os.environ.get("CONDA_EXE"):
            return os.environ["CONDA_EXE"]

        executables = [f"{conda_executable}.exe", conda_executable]
        for exe in executables:
            path = shutil.which(exe)
            if path is not None:
                return path

        print(
            "🔍 Going to search in different common paths"
            f" because `{conda_executable}` was not found in PATH.",
        )
        return _find_windows_path(conda_executable)
    executable = _get_conda_executable(conda_executable)
    assert executable is not None
    return executable


def _capitalize_dir(path: str, *, capitalize: bool = True, index: int = -1) -> str:
    """Capitalize or lowercase a directory in a path, on Windows only."""
    sep = "\\"
    parts = path.split(sep)
    if capitalize:
        parts[index] = parts[index].capitalize()
    else:
        parts[index] = parts[index].lower()
    return sep.join(parts)


@functools.lru_cache(1)
def _find_windows_path(conda_executable: CondaExecutable) -> str:
    """Find the path to the conda executable on Windows."""
    searched = []
    conda_roots = [
        r"%USERPROFILE%\Anaconda3",  # https://stackoverflow.com/a/58211115
        r"%USERPROFILE%\Miniconda3",  # https://stackoverflow.com/a/76545804
        r"C:\Anaconda3",  # https://stackoverflow.com/a/44597801
        r"C:\Miniconda3",  # https://stackoverflow.com/a/53685910
        r"C:\ProgramData\Anaconda3",  # https://stackoverflow.com/a/58211115
        r"C:\ProgramData\Miniconda3",  # https://stackoverflow.com/a/51003321
    ]
    if conda_executable == "mamba":
        conda_roots = [
            r"C:\ProgramData\mambaforge",  # https://github.com/mamba-org/mamba/issues/1756#issuecomment-1517284831
            r"%USERPROFILE%\AppData\Local\mambaforge",  # https://stackoverflow.com/a/75612393
            # First try native mamba locations, then Conda locations (in
            # case `conda install mamba` was used)
            *conda_roots,
        ]
    if conda_executable == "micromamba":
        conda_roots = [
            # Default installation directory based on the installation script
            # https://raw.githubusercontent.com/mamba-org/micromamba-releases/main/install.ps1
            r"%LOCALAPPDATA%\micromamba",
        ]

    extensions = (".exe", "", ".bat")
    subs = ("condabin\\", "Scripts\\", "")  # The "" is for micromamba
    for root, sub, ext, cap in itertools.product(
        conda_roots,
        subs,
        extensions,
        (True, False),
    ):
        # @sbalk reported that his `anaconda3` folder is lowercase
        path = rf"{_capitalize_dir(root, capitalize=cap)}\{sub}{conda_executable}{ext}"
        path = os.path.expandvars(path)
        searched.append(path)
        if os.path.exists(path):  # noqa: PTH110
            return path
    msg = f"Could not find {conda_executable}."
    searched_str = "\n👉 ".join(searched)
    msg = f"Could not find {conda_executable}. Searched in:\n👉 {searched_str}"
    raise FileNotFoundError(msg)


def _conda_cli_command_json(
    conda_executable: CondaExecutable,
    *args: str,
) -> dict[str, list[str]]:
    """Run a conda command and return the JSON output."""
    try:
        result = subprocess.run(
            [_maybe_exe(conda_executable), *args, "--json"],
            capture_output=True,
            text=True,
            check=True,
        )
        return json.loads(result.stdout)
    except subprocess.CalledProcessError as e:  # pragma: no cover
        print(f"Error occurred: {e}")
        raise
    except json.JSONDecodeError as e:  # pragma: no cover
        print(f"Failed to parse JSON: {e}")
        raise


@functools.lru_cache(maxsize=None)
def _conda_env_list(conda_executable: CondaExecutable) -> list[str]:
    """Get a list of conda environments."""
    return _conda_cli_command_json(conda_executable, "env", "list")["envs"]


@functools.lru_cache(maxsize=None)
def _conda_info(conda_executable: CondaExecutable) -> dict:
    return _conda_cli_command_json(conda_executable, "info")


def _conda_root_prefix(conda_executable: CondaExecutable) -> Path:  # pragma: no cover
    """Get the root prefix of the conda installation."""
    if os.environ.get("MAMBA_ROOT_PREFIX"):
        return Path(os.environ["MAMBA_ROOT_PREFIX"])
    if os.environ.get("CONDA_ROOT"):
        return Path(os.environ["CONDA_ROOT"])
    info_dict = _conda_info(conda_executable)
    if conda_executable in ("conda", "mamba"):
        prefix = info_dict.get("root_prefix") or info_dict["conda_prefix"]
    else:
        assert conda_executable == "micromamba"
        prefix = info_dict["base environment"]
    return Path(prefix)


def _conda_env_dirs(
    conda_executable: CondaExecutable,
) -> list[Path]:  # pragma: no cover
    """Get a list of conda environment directories."""
    info_dict = _conda_info(conda_executable)
    if conda_executable in ("conda", "mamba"):
        envs_dirs = info_dict["envs_dirs"]
    else:
        assert conda_executable == "micromamba"
        envs_dirs = info_dict["envs directories"]
    return [Path(d) for d in envs_dirs]


def _conda_env_name_to_prefix(
    conda_executable: CondaExecutable,
    conda_env_name: str,
    *,
    raise_if_not_found: bool = True,
) -> Path | None:  # pragma: no cover
    """Get the prefix of a conda environment."""
    # Based on `conda.base.context.locate_prefix_by_name`
    # https://github.com/conda/conda/blob/72fe69dac8b2fef351c511c813493fef17f295e1/conda/base/context.py#L1976-L1977
    root_prefix = _conda_root_prefix(conda_executable)
    if conda_env_name in ("base", "root"):
        return root_prefix

    for envs_dir in _conda_env_dirs(conda_executable):
        prefix = envs_dir / conda_env_name
        if prefix.exists():
            return prefix
    if not raise_if_not_found:
        return None
    envs = _conda_env_list(conda_executable)
    envs_str = "\n👉 ".join(envs)
    msg = (
        f"Could not find conda prefix with name `{conda_env_name}`."
        f" Available prefixes:\n👉 {envs_str}"
    )
    raise ValueError(msg)


def _maybe_create_conda_env_args(
    conda_executable: CondaExecutable,
    conda_env_name: str | None,
    conda_env_prefix: Path | None,
) -> list[str]:
    if not conda_env_name and not conda_env_prefix:
        return []
    conda_env_args = []
    if conda_env_name:
        conda_env_args = ["--name", conda_env_name]
        prefix = _conda_env_name_to_prefix(
            conda_executable,
            conda_env_name,
            raise_if_not_found=False,
        )
        if prefix is None:
            _create_conda_environment(conda_executable, *conda_env_args)
    elif conda_env_prefix:
        conda_env_args = ["--prefix", str(conda_env_prefix)]
        if not conda_env_prefix.exists():
            _create_conda_environment(conda_executable, *conda_env_args)
    return conda_env_args


def _create_conda_environment(
    conda_executable: CondaExecutable,
    *args: str,
) -> None:  # pragma: no cover
    """Create an empty conda environment."""
    conda_command = [_maybe_exe(conda_executable), "create", "--yes", *args]
    print(f"📦 Creating empty conda environment with `{' '.join(conda_command)}`\n")
    subprocess.run(conda_command, check=True)


def _python_executable(
    conda_executable: CondaExecutable | None,
    conda_env_name: str | None,
    conda_env_prefix: Path | None,
) -> str:
    """Get the Python executable to use for a conda environment."""
    if conda_env_name is None and conda_env_prefix is None:
        return sys.executable
    if conda_env_name:
        assert conda_executable is not None
        conda_env_prefix = _conda_env_name_to_prefix(conda_executable, conda_env_name)
    assert conda_env_prefix is not None
    if platform.system() == "Windows":  # pragma: no cover
        python_executable = conda_env_prefix / "python.exe"
    else:
        python_executable = conda_env_prefix / "bin" / "python"
    assert python_executable.exists()
    return str(python_executable)


def _use_uv(no_uv: bool) -> bool:  # noqa: FBT001
    """Check if the user wants to use the `uv` package."""
    if no_uv:
        return False
    return shutil.which("uv") is not None


def _pip_install_local(
    *folders: str | Path,
    editable: bool,
    dry_run: bool,
    python_executable: str,
    conda_run: list[str],
    no_uv: bool,
    flags: list[str] | None = None,
) -> None:  # pragma: no cover
    if _use_uv(no_uv):
        pip_command = [
            *conda_run,
            "uv",
            "pip",
            "install",
            "--python",
            python_executable,
        ]
    else:
        pip_command = [*conda_run, python_executable, "-m", "pip", "install"]

    if flags:
        pip_command.extend(flags)

    for folder in sorted(folders):
        if not os.path.isabs(folder):  # noqa: PTH117
            relative_prefix = ".\\" if os.name == "nt" else "./"
            folder = f"{relative_prefix}{folder}"  # noqa: PLW2901

        if (
            editable
            and not str(folder).endswith(".whl")
            and not str(folder).endswith(".zip")
        ):
            pip_command.extend(["-e", str(folder)])
        else:
            pip_command.append(str(folder))

    print(f"📦 Installing project with `{' '.join(pip_command)}`\n")
    if not dry_run:
        subprocess.run(pip_command, check=True)


def _install_command(  # noqa: PLR0912, PLR0915
    *files: Path,
    conda_executable: CondaExecutable | None,
    conda_env_name: str | None,
    conda_env_prefix: Path | None,
    conda_lock_file: Path | None,
    dry_run: bool,
    editable: bool,
    skip_local: bool = False,
    skip_pip: bool = False,
    skip_conda: bool = False,
    no_dependencies: bool = False,
    ignore_pins: list[str] | None = None,
    overwrite_pins: list[str] | None = None,
    skip_dependencies: list[str] | None = None,
    no_uv: bool = True,
    verbose: bool = False,
) -> None:
    """Install the dependencies of a single `requirements.yaml` or `pyproject.toml` file."""  # noqa: E501
    start_time = time.time()
    paths_with_extras = [parse_folder_or_filename(f) for f in files]
    requirements = parse_requirements(
        *[f.path for f in paths_with_extras],
        ignore_pins=ignore_pins,
        overwrite_pins=overwrite_pins,
        skip_dependencies=skip_dependencies,
        verbose=verbose,
        extras=[f.extras for f in paths_with_extras],
    )
    platforms = [identify_current_platform()]
    resolved = resolve_conflicts(
        requirements.requirements,
        platforms,
        optional_dependencies=requirements.optional_dependencies,
    )
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms=platforms,
    )
    if not conda_executable:  # None or empty string
        conda_executable = _maybe_conda_executable()
    if conda_lock_file:  # As late as possible to error out early in previous steps
        assert conda_executable is not None
        _create_env_from_lock(
            conda_lock_file,
            conda_executable,
            conda_env_name=conda_env_name,
            conda_env_prefix=conda_env_prefix,
            dry_run=dry_run,
            verbose=verbose,
        )
        no_dependencies = True  # Assume the lock file has all dependencies

    if no_dependencies:
        skip_pip = True
        skip_conda = True

    if env_spec.conda and not skip_conda:
        assert conda_executable is not None
        channel_args = ["--override-channels"] if env_spec.channels else []
        for channel in env_spec.channels:
            channel_args.extend(["--channel", channel])
        conda_env_args = _maybe_create_conda_env_args(
            conda_executable,
            conda_env_name,
            conda_env_prefix,
        )
        conda_command = [
            _maybe_exe(conda_executable),
            "install",
            "--yes",
            *channel_args,
            *conda_env_args,
        ]
        # When running the command in terminal, we need to wrap the pin in quotes
        # so what we print is what the user would type (copy-paste).
        to_print = [_format_inline_conda_package(pkg) for pkg in env_spec.conda]  # type: ignore[arg-type]
        conda_command_str = " ".join((*conda_command, *to_print))
        print(f"📦 Installing conda dependencies with `{conda_command_str}`\n")  # type: ignore[arg-type]
        if not dry_run:  # pragma: no cover
            subprocess.run((*conda_command, *env_spec.conda), check=True)  # type: ignore[arg-type]
    python_executable = _python_executable(
        conda_executable,
        conda_env_name,
        conda_env_prefix,
    )
    if env_spec.pip and not skip_pip:
        conda_run = _maybe_conda_run(conda_executable, conda_env_name, conda_env_prefix)
        if _use_uv(no_uv):
            pip_command = [
                *conda_run,
                "uv",
                "pip",
                "install",
                "--python",
                python_executable,
                *env_spec.pip,
            ]
        else:
            pip_command = [
                *conda_run,
                python_executable,
                "-m",
                "pip",
                "install",
                *env_spec.pip,
            ]
        print(f"📦 Installing pip dependencies with `{' '.join(pip_command)}`\n")
        if not dry_run:  # pragma: no cover
            subprocess.run(pip_command, check=True)

    installable = []
    if not skip_local:
        for file in paths_with_extras:
            if is_pip_installable(file.path.parent):
                installable.append(file.path.parent)
            else:  # pragma: no cover
                print(
                    f"⚠️  Project {file.path.parent} is not pip installable. "
                    "Could not find setup.py or [build-system] in pyproject.toml.",
                )

        # Install local dependencies (if any) included via `local_dependencies:`
        local_dependencies = parse_local_dependencies(
            *[p.path_with_extras for p in paths_with_extras],
            check_pip_installable=True,
            verbose=verbose,
        )
        names = {k.name: [dep.name for dep in v] for k, v in local_dependencies.items()}
        print(f"📝 Found local dependencies: {names}\n")
        installable_set = {p.resolve() for p in installable}
        installable += [
            dep
            for deps in local_dependencies.values()
            for dep in deps
            if dep.resolve() not in installable_set
        ]
        if installable:
            pip_flags = ["--no-deps"]  # we just ran pip/conda install, so skip
            if verbose:
                pip_flags.append("--verbose")
            conda_run = _maybe_conda_run(
                conda_executable,
                conda_env_name,
                conda_env_prefix,
            )
            _pip_install_local(
                *sorted(installable),
                editable=editable,
                dry_run=dry_run,
                python_executable=python_executable,
                flags=pip_flags,
                no_uv=no_uv,
                conda_run=conda_run,
            )

    if not dry_run:  # pragma: no cover
        total_time = time.time() - start_time
        msg = f"✅ All dependencies installed successfully in {total_time:.2f} seconds."
        print(msg)


def _install_all_command(
    *,
    conda_executable: CondaExecutable | None,
    conda_env_name: str | None,
    conda_env_prefix: Path | None,
    conda_lock_file: Path | None,
    dry_run: bool,
    editable: bool,
    depth: int,
    directory: Path,
    skip_local: bool = False,
    skip_pip: bool = False,
    skip_conda: bool = False,
    no_dependencies: bool = False,
    ignore_pins: list[str] | None = None,
    overwrite_pins: list[str] | None = None,
    skip_dependencies: list[str] | None = None,
    no_uv: bool = True,
    verbose: bool = False,
) -> None:  # pragma: no cover
    found_files = find_requirements_files(
        directory,
        depth,
        verbose=verbose,
    )
    if not found_files:
        print(f"❌ No {_DEP_FILES} files found in {directory}")
        sys.exit(1)
    _install_command(
        *found_files,
        conda_executable=conda_executable,
        conda_env_name=conda_env_name,
        conda_env_prefix=conda_env_prefix,
        conda_lock_file=conda_lock_file,
        dry_run=dry_run,
        editable=editable,
        skip_local=skip_local,
        skip_pip=skip_pip,
        skip_conda=skip_conda,
        no_dependencies=no_dependencies,
        ignore_pins=ignore_pins,
        overwrite_pins=overwrite_pins,
        skip_dependencies=skip_dependencies,
        no_uv=no_uv,
        verbose=verbose,
    )


def _maybe_conda_run(
    conda_executable: CondaExecutable | None,
    conda_env_name: str | None,
    conda_env_prefix: Path | None,
) -> list[str]:
    if not conda_executable:  # None or empty string
        return []
    if conda_env_name is None and conda_env_prefix is None:
        if not os.getenv("CONDA_PREFIX") and not os.getenv("MAMBA_ROOT_PREFIX"):
            # Conda/mamba/micromamba might be installed but not in PATH
            return []
        exe = Path(sys.executable)
        conda_prefix = exe.parent if os.name == "nt" else exe.parent.parent
        env_args = ["--prefix", str(conda_prefix)]
    elif conda_env_name:
        env_args = ["--name", conda_env_name]
    elif conda_env_prefix:
        env_args = ["--prefix", str(conda_env_prefix)]
    return [_maybe_exe(conda_executable), "run", *env_args]


def _create_env_from_lock(  # noqa: PLR0912
    conda_lock_file: Path,
    conda_executable: CondaExecutable,
    conda_env_name: str | None,
    conda_env_prefix: Path | None,
    *,
    dry_run: bool,
    verbose: bool,
) -> None:
    if conda_env_name is None and conda_env_prefix is None:
        print(
            "❌ Please provide either `--conda-env-name` or"
            " `--conda-env-prefix` when using `--conda-lock-file`.",
        )
        sys.exit(1)
    elif conda_env_name:
        env_args = ["--name", conda_env_name]
    elif conda_env_prefix:
        env_args = ["--prefix", str(conda_env_prefix)]

    if conda_executable == "micromamba":
        create_cmd = [
            _maybe_exe(conda_executable),
            "create",
            "-f",
            str(conda_lock_file),
            "--yes",
            *env_args,
        ]
        if verbose:
            create_cmd.append("--verbose")
    else:  # conda or mamba
        if not dry_run:
            _verify_conda_lock_installed()
        create_cmd = ["conda-lock", "install", *env_args]

        if conda_executable == "mamba":
            create_cmd.append("--mamba")
        elif conda_executable == "conda":
            create_cmd.extend(["--conda", "conda"])

        create_cmd.append(str(conda_lock_file))

        if verbose:
            create_cmd.append("--log-level=DEBUG")
    create_cmd_str = " ".join(map(str, create_cmd))
    env_identifier = (
        f"'{conda_env_name}'" if conda_env_name else f"at '{conda_env_prefix}'"
    )
    print(f"📦 Creating conda environment {env_identifier} with `{create_cmd_str}`")

    if not dry_run:  # pragma: no cover
        try:
            subprocess.run(create_cmd, check=True)
            if verbose:
                print(f"✅ Environment {env_identifier} created successfully.")
        except subprocess.CalledProcessError as e:
            print(f"❌ Failed to create environment: {e}")
            sys.exit(1)
    else:
        print("🏁 Dry run completed. No environment was created.")


def _verify_conda_lock_installed() -> None:
    """Verify that conda-lock is installed and accessible."""
    if shutil.which("conda-lock") is None:
        print(
            "❌ conda-lock is not installed or not found in PATH.\n"
            "Please install it with one of the following commands:\n"
            "  pip install conda-lock\n"
            "  conda install conda-lock -c conda-forge\n"
            "  mamba install conda-lock -c conda-forge",
        )
        sys.exit(1)

    try:
        # Check if conda-lock is working correctly
        subprocess.run(
            ["conda-lock", "--version"],  # noqa: S607
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        print(
            "❌ conda-lock is installed but not working correctly.\n"
            "Please try reinstalling it with one of the following commands:\n"
            "  `pip install --force-reinstall conda-lock`\n"
            "  `conda install --force-reinstall conda-lock -c conda-forge`\n"
            "  `mamba install --force-reinstall conda-lock -c conda-forge`\n"
            "  `pipx install --force-reinstall conda-lock`",
        )
        sys.exit(1)

    # If we get here, conda-lock is installed and working


def _merge_command(
    *,
    depth: int,
    directory: Path,
    files: list[Path] | None,
    name: str,
    output: Path,
    stdout: bool,
    selector: Literal["sel", "comment"],
    platforms: list[Platform],
    ignore_pins: list[str],
    skip_dependencies: list[str],
    overwrite_pins: list[str],
    verbose: bool,
) -> None:  # pragma: no cover
    # When using stdout, suppress verbose output
    verbose = verbose and not stdout

    if files:  # ignores depth and directory!
        found_files = files
    else:
        found_files = find_requirements_files(
            directory,
            depth,
            verbose=verbose,
        )
        if not found_files:
            print(f"❌ No {_DEP_FILES} files found in {directory}")
            sys.exit(1)

    requirements = parse_requirements(
        *found_files,
        ignore_pins=ignore_pins,
        overwrite_pins=overwrite_pins,
        skip_dependencies=skip_dependencies,
        verbose=verbose,
    )
    if not platforms:
        platforms = requirements.platforms or [identify_current_platform()]
    resolved = resolve_conflicts(
        requirements.requirements,
        platforms,
        optional_dependencies=requirements.optional_dependencies,
    )
    env_spec = create_conda_env_specification(
        resolved,
        requirements.channels,
        platforms,
        selector=selector,
    )
    output_file = None if stdout else output
    write_conda_environment_file(env_spec, output_file, name, verbose=verbose)
    if output_file:
        found_files_str = ", ".join(f"`{f}`" for f in found_files)
        print(
            f"✅ Generated environment file at `{output_file}` from {found_files_str}",
        )


def _pip_compile_command(
    *,
    depth: int,
    directory: Path,
    platform: Platform,
    ignore_pins: list[str],
    skip_dependencies: list[str],
    overwrite_pins: list[str],
    verbose: bool,
    extra_flags: list[str],
    output_file: Path | None = None,
) -> None:
    if importlib.util.find_spec("piptools") is None:  # pragma: no cover
        print(
            "❌ Could not import `pip-tools` module."
            " Please install it with `pip install pip-tools`.",
        )
        sys.exit(1)

    found_files = find_requirements_files(
        directory,
        depth,
        verbose=verbose,
    )

    requirements = parse_requirements(
        *found_files,
        ignore_pins=ignore_pins,
        overwrite_pins=overwrite_pins,
        skip_dependencies=skip_dependencies,
        verbose=verbose,
    )
    resolved = resolve_conflicts(
        requirements.requirements,
        [platform],
        optional_dependencies=requirements.optional_dependencies,
    )
    python_deps = filter_python_dependencies(resolved)
    requirements_in = directory / "requirements.in"
    with requirements_in.open("w") as f:
        f.write("\n".join(python_deps))
    print("✅ Generated `requirements.in` file.")
    if extra_flags:
        assert extra_flags[0] == "--"
        extra_flags = extra_flags[1:]
        if verbose:
            print(f"📝 Extra flags for `pip-compile`: {extra_flags}")

    if output_file is None:
        output_file = directory / "requirements.txt"

    cmd = [
        "pip-compile",
        "--output-file",
        str(output_file),
        *extra_flags,
        str(requirements_in),
    ]
    print(f"🔒 Locking dependencies with `{' '.join(cmd)}`\n")
    subprocess.run(cmd, check=True)
    if output_file.exists():  # pragma: no cover
        # might not exist in tests
        add_comment_to_file(output_file)
    print(f"✅ Generated `{output_file}`.")


def _check_conda_prefix() -> None:  # pragma: no cover
    """Check if sys.executable is in the $CONDA_PREFIX."""
    if "CONDA_PREFIX" not in os.environ:
        return
    conda_prefix = os.environ["CONDA_PREFIX"]
    if sys.executable.startswith(str(conda_prefix)):
        return
    msg = (
        "UniDep should be run from the current Conda environment for correct"
        " operation. However, it's currently running with the Python interpreter"
        f" at `{sys.executable}`, which is not in the active Conda environment"
        f" (`{conda_prefix}`). Please install and run UniDep in the current"
        " Conda environment to avoid any issues, or provide the `--conda-env-name`"
        " or `--conda-env-prefix` option to specify the Conda environment to use."
    )
    warn(msg, stacklevel=2)
    sys.exit(1)


def _print_versions() -> None:  # pragma: no cover
    """Print version information."""
    path = Path(__file__).parent
    txt = [
        f"unidep version: {__version__}",
        f"unidep location: {path}",
        f"Python version: {sys.version}",
        f"Python executable: {sys.executable}",
    ]
    extra_packages = [
        "rich_argparse",
        "rich",
        "conda_lock",
        "pydantic",
        "pip_tools",
        "conda_package_handling",
        "ruamel.yaml",
        "packaging",
        "tomli",
    ]
    for package in extra_packages:
        version = get_package_version(package)
        if version is not None:
            txt.append(f"{package} version: {version}")

    if importlib.util.find_spec("rich") is not None:
        _print_with_rich(txt)
    else:
        print("\n".join(txt))


def _print_with_rich(data: list) -> None:
    """Print data as a table using rich, if it's installed."""
    from rich.console import Console
    from rich.table import Table

    console = Console()
    table = Table(show_header=False)
    table.add_column("Property", style="cyan")
    table.add_column("Value", style="magenta")
    for line in data:
        prop, value = line.split(":", 1)
        table.add_row(prop, value.strip())
    console.print(table)


def _pip_subcommand(
    *,
    file: list[Path],
    platforms: list[Platform],
    verbose: bool,
    ignore_pins: list[str] | None,
    skip_dependencies: list[str] | None,
    overwrite_pins: list[str] | None,
    separator: str,
) -> str:  # pragma: no cover
    platforms = platforms or [identify_current_platform()]
    assert len(file) <= 1
    path = file[0] if file else Path()
    deps = get_python_dependencies(
        path,
        platforms=platforms,
        verbose=verbose,
        ignore_pins=ignore_pins,
        skip_dependencies=skip_dependencies,
        overwrite_pins=overwrite_pins,
        include_local_dependencies=True,
    )
    pip_dependencies = deps.dependencies
    for extra in parse_folder_or_filename(path).extras:
        pip_dependencies.extend(deps.extras[extra])
    return escape_unicode(separator).join(pip_dependencies)


def main() -> None:
    """Main entry point for the command-line tool."""
    args = _parse_args()

    if args.command == "merge":  # pragma: no cover
        _merge_command(
            depth=args.depth,
            directory=args.directory,
            files=None,
            name=args.name,
            output=args.output,
            stdout=args.stdout,
            selector=args.selector,
            platforms=args.platform,
            ignore_pins=args.ignore_pin,
            skip_dependencies=args.skip_dependency,
            overwrite_pins=args.overwrite_pin,
            verbose=args.verbose,
        )
    elif args.command == "pip":  # pragma: no cover
        print(
            _pip_subcommand(
                file=args.file,
                platforms=args.platform,
                verbose=args.verbose,
                ignore_pins=args.ignore_pin,
                skip_dependencies=args.skip_dependency,
                overwrite_pins=args.overwrite_pin,
                separator=args.separator,
            ),
        )
    elif args.command == "conda":  # pragma: no cover
        platforms = args.platform or [identify_current_platform()]
        files = args.file or [Path()]
        requirements = parse_requirements(
            *files,
            ignore_pins=args.ignore_pin,
            skip_dependencies=args.skip_dependency,
            overwrite_pins=args.overwrite_pin,
            verbose=args.verbose,
        )
        resolved = resolve_conflicts(
            requirements.requirements,
            platforms,
            optional_dependencies=requirements.optional_dependencies,
        )
        env_spec = create_conda_env_specification(
            resolved,
            requirements.channels,
            platforms=platforms,
        )

        if any(parse_folder_or_filename(f).extras for f in files):
            msg = "🚧 The `conda` command currently does not support extras."
            print(msg)
            sys.exit(1)

        print(escape_unicode(args.separator).join(env_spec.conda))  # type: ignore[arg-type]
    elif args.command == "install":
        if args.conda_env_name is None and args.conda_env_prefix is None:
            _check_conda_prefix()
        _install_command(
            *(args.files or [Path()]),
            conda_executable=args.conda_executable,
            conda_env_name=args.conda_env_name,
            conda_env_prefix=args.conda_env_prefix,
            conda_lock_file=args.conda_lock_file,
            dry_run=args.dry_run,
            editable=args.editable,
            skip_local=args.skip_local,
            skip_pip=args.skip_pip,
            skip_conda=args.skip_conda,
            no_dependencies=args.no_dependencies,
            ignore_pins=args.ignore_pin,
            skip_dependencies=args.skip_dependency,
            overwrite_pins=args.overwrite_pin,
            no_uv=args.no_uv,
            verbose=args.verbose,
        )
    elif args.command == "install-all":
        if args.conda_env_name is None and args.conda_env_prefix is None:
            _check_conda_prefix()
        _install_all_command(
            conda_executable=args.conda_executable,
            conda_env_name=args.conda_env_name,
            conda_env_prefix=args.conda_env_prefix,
            conda_lock_file=args.conda_lock_file,
            dry_run=args.dry_run,
            editable=args.editable,
            depth=args.depth,
            directory=args.directory,
            skip_local=args.skip_local,
            skip_pip=args.skip_pip,
            skip_conda=args.skip_conda,
            no_dependencies=args.no_dependencies,
            ignore_pins=args.ignore_pin,
            skip_dependencies=args.skip_dependency,
            overwrite_pins=args.overwrite_pin,
            no_uv=args.no_uv,
            verbose=args.verbose,
        )
    elif args.command == "conda-lock":  # pragma: no cover
        conda_lock_command(
            depth=args.depth,
            directory=args.directory,
            files=args.file or None,
            platforms=args.platform,
            verbose=args.verbose,
            only_global=args.only_global,
            ignore_pins=args.ignore_pin,
            skip_dependencies=args.skip_dependency,
            overwrite_pins=args.overwrite_pin,
            check_input_hash=args.check_input_hash,
            extra_flags=args.extra_flags,
            lockfile=args.lockfile,
        )
    elif args.command == "pip-compile":  # pragma: no cover
        if args.platform and len(args.platform) > 1:
            print(
                "❌ The `pip-compile` command does not support multiple platforms.",
            )
            sys.exit(1)
        platform = args.platform[0] if args.platform else identify_current_platform()
        _pip_compile_command(
            depth=args.depth,
            directory=args.directory,
            platform=platform,
            verbose=args.verbose,
            ignore_pins=args.ignore_pin,
            skip_dependencies=args.skip_dependency,
            overwrite_pins=args.overwrite_pin,
            extra_flags=args.extra_flags,
            output_file=args.output_file,
        )
    elif args.command == "version":  # pragma: no cover
        _print_versions()



"""Tests for the `project_dependency_handling` feature."""

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING, Literal

import pytest

from unidep._dependencies_parsing import (
    _add_project_dependencies,
    parse_requirements,
)
from unidep.platform_definitions import Spec

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("project_dependencies", "handling_mode", "expected"),
    [
        # Test same-name
        (
            ["pandas", "requests"],
            "same-name",
            ["pandas", "requests"],
        ),
        # Test pip-only
        (
            ["pandas", "requests"],
            "pip-only",
            [{"pip": "pandas"}, {"pip": "requests"}],
        ),
        # Test ignore
        (["pandas", "requests"], "ignore", []),
        # Test invalid handling mode
        (["pandas", "requests"], "invalid", []),
    ],
)
def test_project_dependency_handling(
    project_dependencies: list[str],
    handling_mode: Literal["same-name", "pip-only", "ignore", "invalid"],
    expected: list[dict[str, str] | str],
) -> None:
    valid_unidep_dependencies: list[dict[str, str] | str] = [
        {"conda": "pandas", "pip": "pandas"},
        "requests",
        {"conda": "zstd", "pip": "zstandard"},
    ]
    unidep_dependencies = valid_unidep_dependencies.copy()
    if handling_mode == "invalid":
        with pytest.raises(ValueError, match="Invalid `project_dependency_handling`"):
            _add_project_dependencies(
                project_dependencies,
                unidep_dependencies,
                handling_mode,  # type: ignore[arg-type]
            )
    else:
        _add_project_dependencies(
            project_dependencies,
            unidep_dependencies,
            handling_mode,  # type: ignore[arg-type]
        )
        assert unidep_dependencies == valid_unidep_dependencies + expected


@pytest.mark.parametrize(
    "project_dependency_handling",
    ["same-name", "pip-only", "ignore"],
)
def test_project_dependency_handling_in_pyproject_toml(
    tmp_path: Path,
    project_dependency_handling: Literal["same-name", "pip-only", "ignore"],
) -> None:
    p = tmp_path / "pyproject.toml"
    p.write_text(
        textwrap.dedent(
            f"""\
            [build-system]
            requires = ["hatchling", "unidep"]
            build-backend = "hatchling.build"

            [project]
            name = "my-project"
            version = "0.1.0"
            dependencies = [
                "requests",
                "pandas",
            ]

            [tool.unidep]
            project_dependency_handling = "{project_dependency_handling}"
            dependencies = [
                {{ conda = "python-graphviz", pip = "graphviz" }},
                {{ conda = "graphviz" }},
            ]
            """,
        ),
    )

    requirements = parse_requirements(p)

    expected = {
        "python-graphviz": [
            Spec(name="python-graphviz", which="conda", identifier="17e5d607"),
        ],
        "graphviz": [
            Spec(name="graphviz", which="pip", identifier="17e5d607"),
            Spec(name="graphviz", which="conda", identifier="5eb93b8c"),
        ],
    }
    if project_dependency_handling == "pip-only":
        expected.update(
            {
                "requests": [Spec(name="requests", which="pip", identifier="08fd8713")],
                "pandas": [Spec(name="pandas", which="pip", identifier="9e467fa1")],
            },
        )
    elif project_dependency_handling == "same-name":
        expected.update(
            {
                "requests": [
                    Spec(name="requests", which="conda", identifier="08fd8713"),
                    Spec(name="requests", which="pip", identifier="08fd8713"),
                ],
                "pandas": [
                    Spec(name="pandas", which="conda", identifier="9e467fa1"),
                    Spec(name="pandas", which="pip", identifier="9e467fa1"),
                ],
            },
        )
    else:
        assert project_dependency_handling == "ignore"
    assert requirements.requirements == expected



#!/usr/bin/env python3
"""unidep - Unified Conda and Pip requirements management.

This module provides setuptools integration for unidep.
"""

from __future__ import annotations

import ast
import configparser
import contextlib
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

from unidep._conflicts import resolve_conflicts
from unidep._dependencies_parsing import parse_local_dependencies, parse_requirements
from unidep.utils import (
    UnsupportedPlatformError,
    build_pep508_environment_marker,
    identify_current_platform,
    parse_folder_or_filename,
    warn,
)

if sys.version_info >= (3, 11):
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib


if TYPE_CHECKING:
    from setuptools import Distribution

    from unidep.platform_definitions import (
        CondaPip,
        Platform,
        Spec,
    )

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal


def filter_python_dependencies(
    resolved: dict[str, dict[Platform | None, dict[CondaPip, Spec]]],
) -> list[str]:
    """Filter out conda dependencies and return only pip dependencies.

    Examples
    --------
    >>> requirements = parse_requirements("requirements.yaml")
    >>> resolved = resolve_conflicts(
    ...     requirements.requirements, requirements.platforms
    ... )
    >>> python_deps = filter_python_dependencies(resolved)

    """
    pip_deps = []
    for platform_data in resolved.values():
        to_process: dict[Platform | None, Spec] = {}  # platform -> Spec
        for _platform, sources in platform_data.items():
            pip_spec = sources.get("pip")
            if pip_spec:
                to_process[_platform] = pip_spec
        if not to_process:
            continue

        # Check if all Spec objects are identical
        first_spec = next(iter(to_process.values()))
        if all(spec == first_spec for spec in to_process.values()):
            # Build a single combined environment marker
            dep_str = first_spec.name_with_pin(is_pip=True)
            if _platform is not None:
                selector = build_pep508_environment_marker(list(to_process.keys()))  # type: ignore[arg-type]
                dep_str = f"{dep_str}; {selector}"
            pip_deps.append(dep_str)
            continue

        for _platform, pip_spec in to_process.items():
            dep_str = pip_spec.name_with_pin(is_pip=True)
            if _platform is not None:
                selector = build_pep508_environment_marker([_platform])
                dep_str = f"{dep_str}; {selector}"
            pip_deps.append(dep_str)
    return sorted(pip_deps)


class Dependencies(NamedTuple):
    dependencies: list[str]
    extras: dict[str, list[str]]


def get_python_dependencies(
    filename: str
    | Path
    | Literal["requirements.yaml", "pyproject.toml"] = "requirements.yaml",  # noqa: PYI051
    *,
    verbose: bool = False,
    ignore_pins: list[str] | None = None,
    overwrite_pins: list[str] | None = None,
    skip_dependencies: list[str] | None = None,
    platforms: list[Platform] | None = None,
    raises_if_missing: bool = True,
    include_local_dependencies: bool = False,
) -> Dependencies:
    """Extract Python (pip) requirements from a `requirements.yaml` or `pyproject.toml` file."""  # noqa: E501
    try:
        p = parse_folder_or_filename(filename)
    except FileNotFoundError:
        if raises_if_missing:
            raise
        return Dependencies(dependencies=[], extras={})

    requirements = parse_requirements(
        p.path,
        ignore_pins=ignore_pins,
        overwrite_pins=overwrite_pins,
        skip_dependencies=skip_dependencies,
        verbose=verbose,
        extras="*",
    )
    if not platforms:
        platforms = list(requirements.platforms)
    resolved = resolve_conflicts(requirements.requirements, platforms)
    dependencies = filter_python_dependencies(resolved)
    # TODO[Bas]: This currently doesn't correctly handle  # noqa: TD004, TD003, FIX002
    # conflicts between sections in the extras and the main dependencies.
    extras = {
        section: filter_python_dependencies(resolve_conflicts(reqs, platforms))
        for section, reqs in requirements.optional_dependencies.items()
    }
    if include_local_dependencies:
        local_dependencies = parse_local_dependencies(
            p.path_with_extras,
            check_pip_installable=True,
            verbose=verbose,
            raise_if_missing=False,  # skip if local dep is not found
            # We don't need to warn about skipping the dependencies of
            # the local dependencies. That is only relevant for the
            # `unidep install` commands.
            warn_non_managed=False,
        )
        for paths in local_dependencies.values():
            for path in paths:
                name = _package_name_from_path(path)
                # TODO: Consider doing this properly using pathname2url  # noqa: FIX002
                # https://github.com/basnijholt/unidep/pull/214#issuecomment-2568663364
                uri = path.as_posix().replace(" ", "%20")
                dependencies.append(f"{name} @ file://{uri}")

    return Dependencies(dependencies=dependencies, extras=extras)


def _package_name_from_setup_cfg(file_path: Path) -> str:
    config = configparser.ConfigParser()
    config.read(file_path)
    name = config.get("metadata", "name", fallback=None)
    if name is None:
        msg = "Could not find the package name in the setup.cfg file."
        raise KeyError(msg)
    return name


def _package_name_from_setup_py(file_path: Path) -> str:
    with file_path.open() as f:
        file_content = f.read()

    tree = ast.parse(file_content)

    class SetupVisitor(ast.NodeVisitor):
        def __init__(self) -> None:
            self.package_name = None

        def visit_Call(self, node: ast.Call) -> None:  # noqa: N802
            if isinstance(node.func, ast.Name) and node.func.id == "setup":
                for keyword in node.keywords:
                    if keyword.arg == "name":
                        self.package_name = keyword.value.value  # type: ignore[attr-defined]

    visitor = SetupVisitor()
    visitor.visit(tree)
    if visitor.package_name is None:
        msg = "Could not find the package name in the setup.py file."
        raise KeyError(msg)
    assert isinstance(visitor.package_name, str)
    return visitor.package_name


def _package_name_from_pyproject_toml(file_path: Path) -> str:
    with file_path.open("rb") as f:
        data = tomllib.load(f)
    with contextlib.suppress(KeyError):
        # PEP 621: setuptools, flit, hatch, pdm
        return data["project"]["name"]
    with contextlib.suppress(KeyError):
        # poetry doesn't follow any standard
        return data["tool"]["poetry"]["name"]
    msg = f"Could not find the package name in the pyproject.toml file: {data}."
    raise KeyError(msg)


def _package_name_from_path(path: Path) -> str:
    """Get the package name from a path."""
    pyproject_toml = path / "pyproject.toml"
    if pyproject_toml.exists():
        with contextlib.suppress(Exception):
            return _package_name_from_pyproject_toml(pyproject_toml)

    setup_cfg = path / "setup.cfg"
    if setup_cfg.exists():
        with contextlib.suppress(Exception):
            return _package_name_from_setup_cfg(setup_cfg)

    setup_py = path / "setup.py"
    if setup_py.exists():
        with contextlib.suppress(Exception):
            return _package_name_from_setup_py(setup_py)

    # Best guess for the package name is folder name.
    return path.name


def _deps(requirements_file: Path) -> Dependencies:  # pragma: no cover
    try:
        platforms = [identify_current_platform()]
    except UnsupportedPlatformError:
        warn(
            "Could not identify the current platform."
            " This may result in selecting all platforms."
            " Please report this issue at"
            " https://github.com/basnijholt/unidep/issues",
        )
        # We don't know the current platform, so we can't filter out.
        # This will result in selecting all platforms. But this is better
        # than failing.
        platforms = None

    skip_local_dependencies = bool(os.getenv("UNIDEP_SKIP_LOCAL_DEPS"))
    verbose = bool(os.getenv("UNIDEP_VERBOSE"))
    return get_python_dependencies(
        requirements_file,
        platforms=platforms,
        raises_if_missing=False,
        verbose=verbose,
        include_local_dependencies=not skip_local_dependencies,
    )


def _setuptools_finalizer(dist: Distribution) -> None:  # pragma: no cover
    """Entry point called by setuptools to get the dependencies for a project."""
    # PEP 517 says that "All hooks are run with working directory set to the
    # root of the source tree".
    project_root = Path.cwd()
    try:
        requirements_file = parse_folder_or_filename(project_root).path
    except FileNotFoundError:
        return
    if requirements_file.exists() and dist.install_requires:  # type: ignore[attr-defined]
        msg = (
            "You have a `requirements.yaml` file in your project root or"
            " configured unidep in `pyproject.toml` with `[tool.unidep]`,"
            " but you are also using setuptools' `install_requires`."
            " Remove the `install_requires` line from `setup.py`."
        )
        raise RuntimeError(msg)

    deps = _deps(requirements_file)
    dist.install_requires = deps.dependencies  # type: ignore[attr-defined]

    if deps.extras:
        dist.extras_require = deps.extras  # type: ignore[attr-defined]



"""unidep CLI tests."""

from __future__ import annotations

import os
import platform
import re
import shutil
import subprocess
import textwrap
from contextlib import contextmanager
from pathlib import Path
from typing import Generator
from unittest.mock import patch

import pytest

from unidep._cli import (
    CondaExecutable,
    _capitalize_dir,
    _conda_env_list,
    _conda_root_prefix,
    _find_windows_path,
    _identify_conda_executable,
    _install_all_command,
    _install_command,
    _maybe_conda_run,
    _maybe_create_conda_env_args,
    _pip_compile_command,
    _pip_subcommand,
    _print_versions,
)

REPO_ROOT = Path(__file__).parent.parent

EXAMPLE_PROJECTS = [
    "setup_py_project",
    "setuptools_project",
    "hatch_project",
    "pyproject_toml_project",
    "hatch2_project",
]


def current_env_and_prefix() -> tuple[str, Path]:
    """Get the current conda environment name and prefix."""
    try:
        prefix = _conda_root_prefix("conda")
    except (KeyError, FileNotFoundError):
        prefix = _conda_root_prefix("micromamba")
    folder, env_name = Path(os.environ["CONDA_PREFIX"]).parts[-2:]
    if folder != "envs":
        return "base", prefix
    return env_name, prefix / "envs" / env_name


@pytest.mark.parametrize(
    "project",
    EXAMPLE_PROJECTS,
)
def test_install_command(project: str, capsys: pytest.CaptureFixture) -> None:
    current_env, prefix = current_env_and_prefix()
    print(f"current_env: {current_env}, prefix: {prefix}")
    for kw in [
        {"conda_env_name": current_env, "conda_env_prefix": None},
        {"conda_env_name": None, "conda_env_prefix": prefix},
    ]:
        _install_command(
            REPO_ROOT / "example" / project,
            conda_executable="",  # type: ignore[arg-type]
            conda_lock_file=None,
            dry_run=True,
            editable=False,
            verbose=True,
            **kw,  # type: ignore[arg-type]
        )
        captured = capsys.readouterr()
        assert "Installing conda dependencies" in captured.out
        assert "Installing pip dependencies" in captured.out
        assert "Installing project with" in captured.out


@pytest.mark.parametrize(
    "project",
    EXAMPLE_PROJECTS,
)
def test_unidep_install_dry_run(project: str) -> None:
    # Path to the requirements file
    requirements_path = REPO_ROOT / "example" / project

    # Ensure the requirements file exists
    assert requirements_path.exists(), "Requirements file does not exist"

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "install",
            "--dry-run",
            str(requirements_path),
        ],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )

    # Check the output
    assert result.returncode == 0, "Command failed to execute successfully"
    if project in ("setup_py_project", "setuptools_project"):
        assert "📦 Installing conda dependencies with" in result.stdout
    assert "📦 Installing pip dependencies with" in result.stdout
    assert "📦 Installing project with" in result.stdout


def test_install_all_command(capsys: pytest.CaptureFixture) -> None:
    _install_all_command(
        conda_executable="",  # type: ignore[arg-type]
        conda_env_name=None,
        conda_env_prefix=None,
        conda_lock_file=None,
        dry_run=True,
        editable=True,
        directory=REPO_ROOT / "example",
        depth=1,
        verbose=False,
    )
    captured = capsys.readouterr()
    assert "Installing conda dependencies" in captured.out
    assert "Installing pip dependencies" in captured.out
    projects = [REPO_ROOT / "example" / p for p in EXAMPLE_PROJECTS]
    pkgs = " ".join([f"-e {p}" for p in sorted(projects)])
    assert f"pip install --no-deps {pkgs}`" in captured.out


def mock_uv_env(tmp_path: Path) -> dict[str, str]:
    """Create a mock uv executable and return env with it in the PATH."""
    mock_uv_path = tmp_path / ("uv.bat" if platform.system() == "Windows" else "uv")
    if platform.system() == "Windows":
        mock_uv_path.write_text("@echo off\necho Mock uv called %*")
    else:
        mock_uv_path.write_text("#!/bin/sh\necho 'Mock uv called' \"$@\"")
    mock_uv_path.chmod(0o755)  # Make it executable

    # Add tmp_path to the PATH environment variable
    env = os.environ.copy()
    env["PATH"] = f"{tmp_path}{os.pathsep}{env['PATH']}"
    return env


@pytest.mark.parametrize("with_uv", [True, False])
def test_unidep_install_all_dry_run(tmp_path: Path, with_uv: bool) -> None:  # noqa: FBT001
    # Path to the requirements file
    requirements_path = REPO_ROOT / "example"

    # Ensure the requirements file exists
    assert requirements_path.exists(), "Requirements file does not exist"

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "install-all",
            "--dry-run",
            "--editable",
            "--directory",
            str(requirements_path),
            *(["--no-uv"] if not with_uv else []),
        ],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
        env=mock_uv_env(tmp_path) if with_uv else None,
    )

    # Check the output
    assert result.returncode == 0, "Command failed to execute successfully"
    assert "📦 Installing conda dependencies with `" in result.stdout

    assert r"📦 Installing pip dependencies with `" in result.stdout
    assert (
        "📝 Found local dependencies: {'pyproject_toml_project': ['hatch_project'], 'setup_py_project': ['hatch_project', 'setuptools_project'], 'setuptools_project': ['hatch_project']}"
        in result.stdout
    )
    projects = [REPO_ROOT / "example" / p for p in EXAMPLE_PROJECTS]
    pkgs = " ".join([f"-e {p}" for p in sorted(projects)])
    assert "📦 Installing project with `" in result.stdout
    if with_uv:
        assert "uv pip install --python" in result.stdout
    else:
        assert f" -m pip install --no-deps {pkgs}" in result.stdout


def test_unidep_conda() -> None:
    # Path to the requirements file
    requirements_path = REPO_ROOT / "example" / "setup_py_project"

    assert requirements_path.exists(), "Requirements file does not exist"

    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "conda",
            "--file",
            str(requirements_path),
        ],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )

    # Check the output
    assert result.returncode == 0, "Command failed to execute successfully"
    assert "pandas" in result.stdout


def test_unidep_file_not_found_error() -> None:
    # Path to the requirements file
    requirements_path = REPO_ROOT / "yolo"

    assert not requirements_path.exists()

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "conda",
            "--file",
            str(requirements_path),
        ],
        check=False,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )

    assert result.returncode == 1, "Command unexpectedly succeeded"
    assert "❌ One or more files" in result.stdout


def test_doubly_nested_project_folder_installable(
    tmp_path: Path,
) -> None:
    example_folder = tmp_path / "example"
    shutil.copytree(REPO_ROOT / "example", example_folder)

    # Add an extra project
    extra_projects = example_folder / "extra_projects"
    extra_projects.mkdir(exist_ok=True, parents=True)
    project4 = extra_projects / "project4"
    project4.mkdir(exist_ok=True, parents=True)
    (project4 / "requirements.yaml").write_text(
        "local_dependencies: [../../setup_py_project]",
    )
    pyproject_toml = "\n".join(  # noqa: FLY002
        (
            "[build-system]",
            'requires = ["setuptools", "unidep"]',
            'build-backend = "setuptools.build_meta"',
        ),
    )

    (project4 / "pyproject.toml").write_text(pyproject_toml)
    setup = "\n".join(  # noqa: FLY002
        (
            "from setuptools import setup",
            'setup(name="project4", version="0.1.0", description="yolo", py_modules=["setup_py_project"])',
        ),
    )
    (project4 / "setup.py").write_text(setup)
    (project4 / "project4.py").write_text("print('hello')")

    # Run the unidep install command
    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "install",
            "--dry-run",
            "--editable",
            "--no-dependencies",
            "--no-uv",
            str(project4 / "requirements.yaml"),
        ],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )

    p1 = str(tmp_path / "example" / "hatch_project")
    p2 = str(tmp_path / "example" / "setup_py_project")
    p3 = str(tmp_path / "example" / "setuptools_project")
    p4 = str(tmp_path / "example" / "extra_projects" / "project4")
    pkgs = " ".join([f"-e {p}" for p in sorted((p1, p2, p3, p4))])
    assert f"pip install --no-deps {pkgs}`" in result.stdout

    p5 = str(tmp_path / "example" / "pyproject_toml_project")
    p6 = str(tmp_path / "example" / "hatch2_project")
    # Test depth 2
    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "install-all",
            "--dry-run",
            "--editable",
            "--no-dependencies",
            "--no-uv",
            "--directory",
            str(example_folder),
            "--depth",
            "2",
        ],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    pkgs = " ".join([f"-e {p}" for p in sorted((p1, p2, p3, p4, p5, p6))])
    assert f"pip install --no-deps {pkgs}`" in result.stdout

    # Test depth 1 (should not install project4)
    result = subprocess.run(
        [  # noqa: S607
            "unidep",
            "install-all",
            "--dry-run",
            "--editable",
            "--no-dependencies",
            "--no-uv",
            "--directory",
            str(example_folder),
            "--depth",
            "1",
        ],
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    pkgs = " ".join([f"-e {p}" for p in sorted((p1, p2, p3, p5, p6))])
    assert f"pip install --no-deps {pkgs}`" in result.stdout


def test_pip_compile_command(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    folder = tmp_path / "example"
    shutil.copytree(REPO_ROOT / "example", folder)
    with patch("subprocess.run", return_value=None), patch(
        "importlib.util.find_spec",
        return_value=True,
    ):
        _pip_compile_command(
            depth=2,
            directory=folder,
            platform="linux-64",
            ignore_pins=[],
            skip_dependencies=[],
            overwrite_pins=[],
            verbose=True,
            extra_flags=["--", "--allow-unsafe"],
        )
    requirements_in = folder / "requirements.in"
    assert requirements_in.exists()
    with requirements_in.open() as f:
        assert "adaptive" in f.read()
    requirements_txt = folder / "requirements.txt"

    assert (
        f"Locking dependencies with `pip-compile --output-file {requirements_txt} --allow-unsafe {requirements_in}`"
        in capsys.readouterr().out
    )


def test_install_non_existing_file() -> None:
    with pytest.raises(FileNotFoundError, match="File `does_not_exist` not found."):
        _install_command(
            Path("does_not_exist"),
            conda_executable="",  # type: ignore[arg-type]
            conda_env_name=None,
            conda_env_prefix=None,
            conda_lock_file=None,
            dry_run=True,
            editable=True,
            verbose=True,
        )


def test_install_non_existing_folder(tmp_path: Path) -> None:
    requirements_file = tmp_path / "requirements.yaml"
    pyproject_file = tmp_path / "pyproject.toml"
    match = re.escape(
        f"File `{requirements_file}` or `{pyproject_file}`"
        f" (with unidep configuration) not found in `{tmp_path}`",
    )
    with pytest.raises(FileNotFoundError, match=match):
        _install_command(
            tmp_path,
            conda_executable="",  # type: ignore[arg-type]
            conda_env_name=None,
            conda_env_prefix=None,
            conda_lock_file=None,
            dry_run=True,
            editable=True,
            verbose=True,
        )


def test_version(capsys: pytest.CaptureFixture) -> None:
    _print_versions()
    captured = capsys.readouterr()
    assert "unidep location" in captured.out
    assert "unidep version" in captured.out
    assert "packaging" in captured.out


def test_conda_env_list() -> None:
    conda_executable = _identify_conda_executable()
    _conda_env_list(conda_executable)


def test_pip_optional(tmp_path: Path) -> None:
    p = tmp_path / "requirements.yaml"
    p.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - foo
            optional_dependencies:
                test:
                    - bar
            """,
        ),
    )
    txt = _pip_subcommand(
        file=[p],
        platforms=[],
        verbose=True,
        ignore_pins=None,
        skip_dependencies=None,
        overwrite_pins=None,
        separator=" ",
    )
    assert txt == "foo"

    txt = _pip_subcommand(
        file=[f"{p}[test]"],  # type: ignore[list-item]
        platforms=[],
        verbose=True,
        ignore_pins=None,
        skip_dependencies=None,
        overwrite_pins=None,
        separator=" ",
    )
    assert txt == "foo bar"


def test_capitalize_last_dir() -> None:
    # Just needs to work for Windows paths
    assert _capitalize_dir(r"foo\bar\baz") == r"foo\bar\Baz"
    assert _capitalize_dir(r"foo\bar\baz", capitalize=False) == r"foo\bar\baz"
    assert _capitalize_dir(r"foo\bar\baz", capitalize=True) == r"foo\bar\Baz"


@pytest.mark.skipif(
    os.name == "nt",
    reason="Don't test on Windows to make sure that conda is not found.",
)
def test_find_conda_windows() -> None:
    """Tests whether the function searches the expected paths."""
    with pytest.raises(
        FileNotFoundError,
        match="Could not find conda.",
    ) as excinfo:
        _find_windows_path("conda")
    # This Windows hell... 🤦‍♂️
    paths = [
        r"👉 %USERPROFILE%\Anaconda3\condabin\conda.exe",
        r"👉 %USERPROFILE%\anaconda3\condabin\conda.exe",
        r"👉 %USERPROFILE%\Anaconda3\condabin\conda",
        r"👉 %USERPROFILE%\anaconda3\condabin\conda",
        r"👉 %USERPROFILE%\Anaconda3\condabin\conda.bat",
        r"👉 %USERPROFILE%\anaconda3\condabin\conda.bat",
        r"👉 %USERPROFILE%\Anaconda3\Scripts\conda.exe",
        r"👉 %USERPROFILE%\anaconda3\Scripts\conda.exe",
        r"👉 %USERPROFILE%\Anaconda3\Scripts\conda",
        r"👉 %USERPROFILE%\anaconda3\Scripts\conda",
        r"👉 %USERPROFILE%\Anaconda3\Scripts\conda.bat",
        r"👉 %USERPROFILE%\anaconda3\Scripts\conda.bat",
        r"👉 %USERPROFILE%\Anaconda3\conda.exe",
        r"👉 %USERPROFILE%\anaconda3\conda.exe",
        r"👉 %USERPROFILE%\Anaconda3\conda",
        r"👉 %USERPROFILE%\anaconda3\conda",
        r"👉 %USERPROFILE%\Anaconda3\conda.bat",
        r"👉 %USERPROFILE%\anaconda3\conda.bat",
        r"👉 %USERPROFILE%\Miniconda3\condabin\conda.exe",
        r"👉 %USERPROFILE%\miniconda3\condabin\conda.exe",
        r"👉 %USERPROFILE%\Miniconda3\condabin\conda",
        r"👉 %USERPROFILE%\miniconda3\condabin\conda",
        r"👉 %USERPROFILE%\Miniconda3\condabin\conda.bat",
        r"👉 %USERPROFILE%\miniconda3\condabin\conda.bat",
        r"👉 %USERPROFILE%\Miniconda3\Scripts\conda.exe",
        r"👉 %USERPROFILE%\miniconda3\Scripts\conda.exe",
        r"👉 %USERPROFILE%\Miniconda3\Scripts\conda",
        r"👉 %USERPROFILE%\miniconda3\Scripts\conda",
        r"👉 %USERPROFILE%\Miniconda3\Scripts\conda.bat",
        r"👉 %USERPROFILE%\miniconda3\Scripts\conda.bat",
        r"👉 %USERPROFILE%\Miniconda3\conda.exe",
        r"👉 %USERPROFILE%\miniconda3\conda.exe",
        r"👉 %USERPROFILE%\Miniconda3\conda",
        r"👉 %USERPROFILE%\miniconda3\conda",
        r"👉 %USERPROFILE%\Miniconda3\conda.bat",
        r"👉 %USERPROFILE%\miniconda3\conda.bat",
        r"👉 C:\Anaconda3\condabin\conda.exe",
        r"👉 C:\anaconda3\condabin\conda.exe",
        r"👉 C:\Anaconda3\condabin\conda",
        r"👉 C:\anaconda3\condabin\conda",
        r"👉 C:\Anaconda3\condabin\conda.bat",
        r"👉 C:\anaconda3\condabin\conda.bat",
        r"👉 C:\Anaconda3\Scripts\conda.exe",
        r"👉 C:\anaconda3\Scripts\conda.exe",
        r"👉 C:\Anaconda3\Scripts\conda",
        r"👉 C:\anaconda3\Scripts\conda",
        r"👉 C:\Anaconda3\Scripts\conda.bat",
        r"👉 C:\anaconda3\Scripts\conda.bat",
        r"👉 C:\Anaconda3\conda.exe",
        r"👉 C:\anaconda3\conda.exe",
        r"👉 C:\Anaconda3\conda",
        r"👉 C:\anaconda3\conda",
        r"👉 C:\Anaconda3\conda.bat",
        r"👉 C:\anaconda3\conda.bat",
        r"👉 C:\Miniconda3\condabin\conda.exe",
        r"👉 C:\miniconda3\condabin\conda.exe",
        r"👉 C:\Miniconda3\condabin\conda",
        r"👉 C:\miniconda3\condabin\conda",
        r"👉 C:\Miniconda3\condabin\conda.bat",
        r"👉 C:\miniconda3\condabin\conda.bat",
        r"👉 C:\Miniconda3\Scripts\conda.exe",
        r"👉 C:\miniconda3\Scripts\conda.exe",
        r"👉 C:\Miniconda3\Scripts\conda",
        r"👉 C:\miniconda3\Scripts\conda",
        r"👉 C:\Miniconda3\Scripts\conda.bat",
        r"👉 C:\miniconda3\Scripts\conda.bat",
        r"👉 C:\Miniconda3\conda.exe",
        r"👉 C:\miniconda3\conda.exe",
        r"👉 C:\Miniconda3\conda",
        r"👉 C:\miniconda3\conda",
        r"👉 C:\Miniconda3\conda.bat",
        r"👉 C:\miniconda3\conda.bat",
        r"👉 C:\ProgramData\Anaconda3\condabin\conda.exe",
        r"👉 C:\ProgramData\anaconda3\condabin\conda.exe",
        r"👉 C:\ProgramData\Anaconda3\condabin\conda",
        r"👉 C:\ProgramData\anaconda3\condabin\conda",
        r"👉 C:\ProgramData\Anaconda3\condabin\conda.bat",
        r"👉 C:\ProgramData\anaconda3\condabin\conda.bat",
        r"👉 C:\ProgramData\Anaconda3\Scripts\conda.exe",
        r"👉 C:\ProgramData\anaconda3\Scripts\conda.exe",
        r"👉 C:\ProgramData\Anaconda3\Scripts\conda",
        r"👉 C:\ProgramData\anaconda3\Scripts\conda",
        r"👉 C:\ProgramData\Anaconda3\Scripts\conda.bat",
        r"👉 C:\ProgramData\anaconda3\Scripts\conda.bat",
        r"👉 C:\ProgramData\Anaconda3\conda.exe",
        r"👉 C:\ProgramData\anaconda3\conda.exe",
        r"👉 C:\ProgramData\Anaconda3\conda",
        r"👉 C:\ProgramData\anaconda3\conda",
        r"👉 C:\ProgramData\Anaconda3\conda.bat",
        r"👉 C:\ProgramData\anaconda3\conda.bat",
        r"👉 C:\ProgramData\Miniconda3\condabin\conda.exe",
        r"👉 C:\ProgramData\miniconda3\condabin\conda.exe",
        r"👉 C:\ProgramData\Miniconda3\condabin\conda",
        r"👉 C:\ProgramData\miniconda3\condabin\conda",
        r"👉 C:\ProgramData\Miniconda3\condabin\conda.bat",
        r"👉 C:\ProgramData\miniconda3\condabin\conda.bat",
        r"👉 C:\ProgramData\Miniconda3\Scripts\conda.exe",
        r"👉 C:\ProgramData\miniconda3\Scripts\conda.exe",
        r"👉 C:\ProgramData\Miniconda3\Scripts\conda",
        r"👉 C:\ProgramData\miniconda3\Scripts\conda",
        r"👉 C:\ProgramData\Miniconda3\Scripts\conda.bat",
        r"👉 C:\ProgramData\miniconda3\Scripts\conda.bat",
        r"👉 C:\ProgramData\Miniconda3\conda.exe",
        r"👉 C:\ProgramData\miniconda3\conda.exe",
        r"👉 C:\ProgramData\Miniconda3\conda",
        r"👉 C:\ProgramData\miniconda3\conda",
        r"👉 C:\ProgramData\Miniconda3\conda.bat",
        r"👉 C:\ProgramData\miniconda3\conda.bat",
    ]
    for path in paths:
        assert path in excinfo.value.args[0]


@contextmanager
def set_env_var(key: str, value: str) -> Generator[None, None, None]:
    original_value = os.environ.get(key)
    os.environ[key] = value
    try:
        yield
    finally:
        if original_value is None:
            del os.environ[key]
        else:
            os.environ[key] = original_value


@pytest.mark.skipif(
    os.name == "nt",
    reason="On Windows it will search for Conda because of `_maybe_exe`.",
)
def test_maybe_conda_run() -> None:
    with set_env_var("CONDA_EXE", "conda"):
        result = _maybe_conda_run("conda", "my_env", None)
        assert result == ["conda", "run", "--name", "my_env"]

    p = Path("/path/to/env")
    with set_env_var("CONDA_EXE", "conda"):
        result = _maybe_conda_run("conda", None, p)
        assert result == ["conda", "run", "--prefix", str(p)]

    with set_env_var("MAMBA_EXE", "mamba"):
        result = _maybe_conda_run("mamba", "my_env", None)
        assert result == ["mamba", "run", "--name", "my_env"]


def test_maybe_create_conda_env_args_creates_env(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
) -> None:
    """Test that _maybe_create_conda_env_args creates the environment if it doesn't exist.

    This simulates running:
      unidep install --conda-env-name non-existing-env .
    and checks that the function to create a conda environment is called.
    """
    # Create a flag to record if _create_conda_environment is called
    created = []

    # Define a fake _create_conda_environment that records its call
    def fake_create(
        conda_executable: CondaExecutable,  # noqa: ARG001
        *args: str,
    ) -> None:
        created.append(args)
        print("Fake create called with", args)

    # Patch the _create_conda_environment function
    monkeypatch.setattr(
        "unidep._cli._create_conda_environment",
        fake_create,
    )

    # Patch _conda_env_name_to_prefix to simulate that the environment is missing.
    def fake_env_name_to_prefix(
        conda_executable: CondaExecutable,  # noqa: ARG001
        env_name: str,  # noqa: ARG001
        *,
        raise_if_not_found: bool = True,  # noqa: ARG001
    ) -> Path | None:
        # Simulate that for "non-existing-env" no environment exists.
        return None

    monkeypatch.setattr(
        "unidep._cli._conda_env_name_to_prefix",
        fake_env_name_to_prefix,
    )

    # Now call _maybe_create_conda_env_args with a non-existing environment name.

    args = _maybe_create_conda_env_args("conda", "non-existing-env", None)

    # Check that our fake_create was called (i.e. the environment creation was triggered)
    assert created, (
        "Expected environment creation to be triggered for non-existing env."
    )
    # Also, the returned arguments should be the standard ones for a named env.
    assert args == ["--name", "non-existing-env"]

    # Optionally, verify that our fake function printed the expected message.
    output = capsys.readouterr().out
    assert "Fake create called with" in output

    # Now with a prefix
    prefix = Path("/home/user/micromamba/envs/non-existing-env")
    args = _maybe_create_conda_env_args("conda", None, prefix)

    # Check that our fake_create was called (i.e. the environment creation was triggered)
    assert created, (
        "Expected environment creation to be triggered for non-existing env."
    )
    # Also, the returned arguments should be the standard ones for a named env.
    assert args == ["--prefix", str(prefix)]

    # Optionally, verify that our fake function printed the expected message.
    output = capsys.readouterr().out
    assert "Fake create called with" in output



"""Tests for setuptools integration."""

import textwrap
from pathlib import Path

import pytest

from unidep._setuptools_integration import (
    _package_name_from_path,
    _package_name_from_pyproject_toml,
    _package_name_from_setup_cfg,
    _package_name_from_setup_py,
)

REPO_ROOT = Path(__file__).parent.parent


def test_package_name_from_path() -> None:
    example = REPO_ROOT / "example"
    # Could not find the package name, so it uses the folder name
    assert _package_name_from_path(example) == "example"
    # The following should read from the setup.py or pyproject.toml file
    assert _package_name_from_path(example / "hatch_project") == "hatch_project"
    assert (
        _package_name_from_pyproject_toml(example / "hatch_project" / "pyproject.toml")
        == "hatch_project"
    )
    assert _package_name_from_path(example / "hatch2_project") == "hatch2_project"
    assert (
        _package_name_from_pyproject_toml(example / "hatch2_project" / "pyproject.toml")
        == "hatch2_project"
    )
    assert (
        _package_name_from_path(example / "pyproject_toml_project")
        == "pyproject_toml_project"
    )
    assert (
        _package_name_from_pyproject_toml(
            example / "pyproject_toml_project" / "pyproject.toml",
        )
        == "pyproject_toml_project"
    )
    assert _package_name_from_path(example / "setup_py_project") == "setup_py_project"
    assert (
        _package_name_from_setup_py(example / "setup_py_project" / "setup.py")
        == "setup_py_project"
    )
    assert (
        _package_name_from_path(example / "setuptools_project") == "setuptools_project"
    )
    assert (
        _package_name_from_pyproject_toml(
            example / "setuptools_project" / "pyproject.toml",
        )
        == "setuptools_project"
    )


def test_package_name_from_cfg(tmp_path: Path) -> None:
    setup_cfg = tmp_path / "setup.cfg"
    setup_cfg.write_text(
        textwrap.dedent(
            """\
            [metadata]
            name = setup_cfg_project
            """,
        ),
    )
    assert _package_name_from_path(tmp_path) == "setup_cfg_project"
    assert _package_name_from_setup_cfg(setup_cfg) == "setup_cfg_project"
    missing = tmp_path / "missing" / "setup.cfg"
    assert not missing.exists()
    with pytest.raises(KeyError):
        _package_name_from_setup_cfg(missing)

    setup_cfg2 = tmp_path / "setup.cfg"
    setup_cfg2.write_text(
        textwrap.dedent(
            """\
            [metadata]
            yolo = missing
            """,
        ),
    )
    with pytest.raises(KeyError):
        _package_name_from_setup_cfg(setup_cfg2)



"""unidep - Unified Conda and Pip requirements management.

This module provides parsing of `requirements.yaml` and `pyproject.toml` files.
"""

from __future__ import annotations

import functools
import hashlib
import os
import sys
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap, CommentedSeq

from unidep.platform_definitions import Platform, Spec, platforms_from_selector
from unidep.utils import (
    PathWithExtras,
    defaultdict_to_dict,
    is_pip_installable,
    parse_folder_or_filename,
    parse_package_str,
    selector_from_comment,
    split_path_and_extras,
    unidep_configured_in_toml,
    warn,
)

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import Literal
    else:  # pragma: no cover
        from typing_extensions import Literal


if sys.version_info >= (3, 11):
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib


def find_requirements_files(
    base_dir: str | Path = ".",
    depth: int = 1,
    *,
    verbose: bool = False,
) -> list[Path]:
    """Scan a directory for `requirements.yaml` and `pyproject.toml` files."""
    base_path = Path(base_dir)
    found_files = []

    # Define a helper function to recursively scan directories
    def _scan_dir(path: Path, current_depth: int) -> None:
        if verbose:
            print(f"🔍 Scanning in `{path}` at depth {current_depth}")
        if current_depth > depth:
            return
        for child in sorted(path.iterdir()):
            if child.is_dir():
                _scan_dir(child, current_depth + 1)
            elif child.name == "requirements.yaml":
                found_files.append(child)
                if verbose:
                    print(f'🔍 Found `"requirements.yaml"` at `{child}`')
            elif child.name == "pyproject.toml" and unidep_configured_in_toml(child):
                if verbose:
                    print(f'🔍 Found `"pyproject.toml"` with dependencies at `{child}`')
                found_files.append(child)

    _scan_dir(base_path, 0)
    return sorted(found_files)


def _extract_first_comment(
    commented_map: CommentedMap,
    index_or_key: int | str,
) -> str | None:
    """Extract the first comment from a CommentedMap."""
    comments = commented_map.ca.items.get(index_or_key, None)
    if comments is None:
        return None
    comment_strings = next(
        c.value.split("\n")[0].rstrip().lstrip() for c in comments if c is not None
    )
    if not comment_strings:
        # empty string
        return None
    return "".join(comment_strings)


def _identifier(identifier: int, selector: str | None) -> str:
    """Return a unique identifier based on the comment."""
    platforms = None if selector is None else tuple(platforms_from_selector(selector))
    data_str = f"{identifier}-{platforms}"
    # Hash using SHA256 and take the first 8 characters for a shorter hash
    return hashlib.sha256(data_str.encode()).hexdigest()[:8]


def _parse_dependency(
    dependency: str,
    dependencies: CommentedMap,
    index_or_key: int | str,
    which: Literal["conda", "pip", "both"],
    identifier: int,
    ignore_pins: list[str],
    overwrite_pins: dict[str, str | None],
    skip_dependencies: list[str],
) -> list[Spec]:
    name, pin, selector = parse_package_str(dependency)
    if name in ignore_pins:
        pin = None
    if name in skip_dependencies:
        return []
    if name in overwrite_pins:
        pin = overwrite_pins[name]
    comment = (
        _extract_first_comment(dependencies, index_or_key)
        if isinstance(dependencies, (CommentedMap, CommentedSeq))
        else None
    )
    if comment and selector is None:
        selector = selector_from_comment(comment)
    identifier_hash = _identifier(identifier, selector)
    if which == "both":
        return [
            Spec(name, "conda", pin, identifier_hash, selector),
            Spec(name, "pip", pin, identifier_hash, selector),
        ]
    return [Spec(name, which, pin, identifier_hash, selector)]


class ParsedRequirements(NamedTuple):
    """Requirements with comments."""

    channels: list[str]
    platforms: list[Platform]
    requirements: dict[str, list[Spec]]
    optional_dependencies: dict[str, dict[str, list[Spec]]]


class Requirements(NamedTuple):
    """Requirements as CommentedSeq."""

    # mypy doesn't support CommentedSeq[str], so we use list[str] instead.
    channels: list[str]  # actually a CommentedSeq[str]
    conda: list[str]  # actually a CommentedSeq[str]
    pip: list[str]  # actually a CommentedSeq[str]


def _parse_overwrite_pins(overwrite_pins: list[str]) -> dict[str, str | None]:
    """Parse overwrite pins."""
    result = {}
    for overwrite_pin in overwrite_pins:
        pkg = parse_package_str(overwrite_pin)
        result[pkg.name] = pkg.pin
    return result


@functools.lru_cache
def _load(p: Path, yaml: YAML) -> dict[str, Any]:
    if p.suffix == ".toml":
        with p.open("rb") as f:
            pyproject = tomllib.load(f)
            project_dependencies = pyproject.get("project", {}).get("dependencies", [])
            unidep_cfg = pyproject["tool"]["unidep"]
            if not project_dependencies:
                return unidep_cfg
            unidep_dependencies = unidep_cfg.setdefault("dependencies", [])
            project_dependency_handling = unidep_cfg.get(
                "project_dependency_handling",
                "ignore",
            )
            _add_project_dependencies(
                project_dependencies,
                unidep_dependencies,
                project_dependency_handling,
            )
            return unidep_cfg
    with p.open() as f:
        return yaml.load(f)


def _add_project_dependencies(
    project_dependencies: list[str],
    unidep_dependencies: list[dict[str, str] | str],
    project_dependency_handling: Literal["same-name", "pip-only", "ignore"],
) -> None:
    """Add project dependencies to unidep dependencies based on the chosen handling."""
    if project_dependency_handling == "same-name":
        unidep_dependencies.extend(project_dependencies)
    elif project_dependency_handling == "pip-only":
        unidep_dependencies.extend([{"pip": dep} for dep in project_dependencies])
    elif project_dependency_handling != "ignore":
        msg = (
            f"Invalid `project_dependency_handling` value: {project_dependency_handling}."  # noqa: E501
            " Must be one of 'same-name', 'pip-only', 'ignore'."
        )
        raise ValueError(msg)


def _get_local_dependencies(data: dict[str, Any]) -> list[str]:
    """Get `local_dependencies` from a `requirements.yaml` or `pyproject.toml` file."""
    if "local_dependencies" in data:
        return data["local_dependencies"]
    if "includes" in data:
        warn(
            "⚠️ You are using `includes` in `requirements.yaml` or `pyproject.toml`"
            " `[unidep.tool]` which is deprecated since 0.42.0 and has been renamed to"
            " `local_dependencies`.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return data["includes"]
    return []


def _to_path_with_extras(
    paths: list[Path],
    extras: list[list[str]] | Literal["*"] | None,
) -> list[PathWithExtras]:
    if isinstance(extras, (list, tuple)) and len(extras) != len(paths):
        msg = (
            f"Length of `extras` ({len(extras)}) does not match length"
            f" of `paths` ({len(paths)})."
        )
        raise ValueError(msg)
    paths_with_extras = [parse_folder_or_filename(p) for p in paths]
    if extras is None:
        return paths_with_extras
    assert extras is not None
    if any(p.extras for p in paths_with_extras):
        msg = (
            "Cannot specify `extras` list when paths are"
            " specified like `path/to/project[extra1,extra2]`, `extras` must be `None`"
            " or specify pure paths without extras like `path/to/project` and specify"
            " extras in `extras`."
        )
        raise ValueError(msg)
    if extras == "*":
        extras = [["*"]] * len(paths)  # type: ignore[list-item]

    return [PathWithExtras(p.path, e) for p, e in zip(paths_with_extras, extras)]


def _update_data_structures(
    *,
    path_with_extras: PathWithExtras,
    datas: list[dict[str, Any]],  # modified in place
    all_extras: list[list[str]],  # modified in place
    seen: set[PathWithExtras],  # modified in place
    yaml: YAML,
    is_nested: bool,
    verbose: bool = False,
) -> None:
    if verbose:
        print(f"📄 Parsing `{path_with_extras.path_with_extras}`")
    data = _load(path_with_extras.path, yaml)
    datas.append(data)
    _move_local_optional_dependencies_to_local_dependencies(
        data=data,  # modified in place
        path_with_extras=path_with_extras,
        verbose=verbose,
    )
    if not is_nested:
        all_extras.append(path_with_extras.extras)
    else:
        # When nested, the extras that are specified in the
        # local_dependencies section should be moved to the main dependencies
        # because they are not optional if specified in the file. Only
        # the top-level extras are optional.
        all_extras.append([])
        _move_optional_dependencies_to_dependencies(
            data=data,  # modified in place
            path_with_extras=path_with_extras,
            verbose=verbose,
        )

    seen.add(path_with_extras.resolved())

    # Handle "local_dependencies" (or old name "includes", changed in 0.42.0)
    for local_dependency in _get_local_dependencies(data):
        # NOTE: The current function calls _add_local_dependencies,
        # which calls the current function recursively
        _add_local_dependencies(
            local_dependency=local_dependency,
            path_with_extras=path_with_extras,
            datas=datas,  # modified in place
            all_extras=all_extras,  # modified in place
            seen=seen,  # modified in place
            yaml=yaml,
            verbose=verbose,
        )


def _move_optional_dependencies_to_dependencies(
    data: dict[str, Any],
    path_with_extras: PathWithExtras,
    *,
    verbose: bool = False,
) -> None:
    optional_dependencies = data.pop("optional_dependencies", {})
    for extra in path_with_extras.extras:
        if extra == "*":
            # If "*" is specified, include all optional dependencies
            for opt_deps in optional_dependencies.values():
                data.setdefault("dependencies", []).extend(opt_deps)
            if verbose:
                print(
                    "📄 Moving all optional dependencies to main dependencies"
                    f" for `{path_with_extras.path_with_extras}`",
                )
        elif extra in optional_dependencies:
            data.setdefault("dependencies", []).extend(optional_dependencies[extra])
            if verbose:
                print(
                    f"📄 Moving `{extra}` optional dependencies to main dependencies"
                    f" for `{path_with_extras.path_with_extras}`",
                )


def _move_local_optional_dependencies_to_local_dependencies(
    *,
    data: dict[str, Any],  # modified in place
    path_with_extras: PathWithExtras,
    verbose: bool = False,
) -> None:
    # Move local dependencies from `optional_dependencies` to `local_dependencies`
    extras = path_with_extras.extras
    if "*" in extras:
        extras = list(data.get("optional_dependencies", {}).keys())

    optional_dependencies = data.get("optional_dependencies", {})
    for extra in extras:
        moved = set()
        for dep in optional_dependencies.get(extra, []):
            if isinstance(dep, dict):
                # This is a {"pip": "package"} and/or {"conda": "package"} dependency
                continue
            if _str_is_path_like(dep):
                if verbose:
                    print(
                        f"📄 Moving `{dep}` from the `{extra}` section in"
                        " `optional_dependencies` to `local_dependencies`",
                    )
                data.setdefault("local_dependencies", []).append(dep)
                moved.add(dep)
        for dep in moved:
            extras = optional_dependencies[extra]  # key must exist if moved non-empty
            extras.pop(extras.index(dep))

    # Remove empty optional_dependencies sections
    to_delete = [extra for extra, deps in optional_dependencies.items() if not deps]
    for extra in to_delete:
        if verbose:
            print(f"📄 Removing empty `{extra}` section from `optional_dependencies`")
        optional_dependencies.pop(extra)


def _add_local_dependencies(
    *,
    local_dependency: str,
    path_with_extras: PathWithExtras,
    datas: list[dict[str, Any]],
    all_extras: list[list[str]],
    seen: set[PathWithExtras],
    yaml: YAML,
    verbose: bool = False,
) -> None:
    try:
        requirements_dep_file = parse_folder_or_filename(
            path_with_extras.path.parent / local_dependency,
        )
    except FileNotFoundError:
        # Means that this is a local package that is not managed by unidep.
        # We do not need to do anything here, just in `unidep install`.
        return
    if requirements_dep_file.path.suffix in (".whl", ".zip"):
        if verbose:
            print(
                f"⚠️  Local dependency `{local_dependency}` is a wheel or zip file. "
                "Skipping parsing, but it will be installed by pip if "
                "`--skip-local` is not set. Note that unidep will not "
                "detect its dependencies.",
            )
        return
    if requirements_dep_file.resolved() in seen:
        return  # Avoids circular local_dependencies
    if verbose:
        print(f"📄 Parsing `{local_dependency}` from `local_dependencies`")
    _update_data_structures(
        path_with_extras=requirements_dep_file,
        datas=datas,  # modified in place
        all_extras=all_extras,  # modified in place
        seen=seen,  # modified in place
        yaml=yaml,
        verbose=verbose,
        is_nested=True,
    )


def parse_requirements(
    *paths: Path,
    ignore_pins: list[str] | None = None,
    overwrite_pins: list[str] | None = None,
    skip_dependencies: list[str] | None = None,
    verbose: bool = False,
    extras: list[list[str]] | Literal["*"] | None = None,
) -> ParsedRequirements:
    """Parse a list of `requirements.yaml` or `pyproject.toml` files.

    Parameters
    ----------
    paths
        Paths to `requirements.yaml` or `pyproject.toml` files.
    ignore_pins
        List of package names to ignore pins for.
    overwrite_pins
        List of package names with pins to overwrite.
    skip_dependencies
        List of package names to skip.
    verbose
        Whether to print verbose output.
    extras
        List of lists of extras to include. The outer list corresponds to the
        `requirements.yaml` or `pyproject.toml` files, the inner list to the
        extras to include for that file. If "*", all extras are included,
        if None, no extras are included.

    """
    paths_with_extras = _to_path_with_extras(paths, extras)  # type: ignore[arg-type]
    ignore_pins = ignore_pins or []
    skip_dependencies = skip_dependencies or []
    overwrite_pins_map = _parse_overwrite_pins(overwrite_pins or [])

    # `data` and `all_extras` are lists of the same length
    datas: list[dict[str, Any]] = []
    all_extras: list[list[str]] = []
    seen: set[PathWithExtras] = set()
    yaml = YAML(typ="rt")  # Might be unused if all are TOML files
    for path_with_extras in paths_with_extras:
        _update_data_structures(
            path_with_extras=path_with_extras,
            datas=datas,  # modified in place
            all_extras=all_extras,  # modified in place
            seen=seen,  # modified in place
            yaml=yaml,
            verbose=verbose,
            is_nested=False,
        )

    assert len(datas) == len(all_extras)

    # Parse the requirements from loaded data
    requirements: dict[str, list[Spec]] = defaultdict(list)
    optional_dependencies: dict[str, dict[str, list[Spec]]] = defaultdict(
        lambda: defaultdict(list),
    )
    channels: set[str] = set()
    platforms: set[Platform] = set()

    identifier = -1
    for data, _extras in zip(datas, all_extras):
        channels.update(data.get("channels", []))
        platforms.update(data.get("platforms", []))
        if "dependencies" in data:
            identifier = _add_dependencies(
                data["dependencies"],
                requirements,  # modified in place
                identifier,
                ignore_pins,
                overwrite_pins_map,
                skip_dependencies,
            )
        for opt_name, opt_deps in data.get("optional_dependencies", {}).items():
            if opt_name in _extras or "*" in _extras:
                identifier = _add_dependencies(
                    opt_deps,
                    optional_dependencies[opt_name],  # modified in place
                    identifier,
                    ignore_pins,
                    overwrite_pins_map,
                    skip_dependencies,
                    is_optional=True,
                )

    return ParsedRequirements(
        sorted(channels),
        sorted(platforms),
        dict(requirements),
        defaultdict_to_dict(optional_dependencies),
    )


def _str_is_path_like(s: str) -> bool:
    """Check if a string is path-like."""
    return os.path.sep in s or "/" in s or s.startswith(".")


def _check_allowed_local_dependency(name: str, is_optional: bool) -> None:  # noqa: FBT001
    if _str_is_path_like(name):
        # There should not be path-like dependencies in the optional_dependencies
        # section after _move_local_optional_dependencies_to_local_dependencies.
        assert not is_optional
        msg = (
            f"Local dependencies (`{name}`) are not allowed in `dependencies`."
            " Use the `local_dependencies` section instead."
        )
        raise ValueError(msg)


def _add_dependencies(
    dependencies: list[str],
    requirements: dict[str, list[Spec]],  # modified in place
    identifier: int,
    ignore_pins: list[str],
    overwrite_pins_map: dict[str, str | None],
    skip_dependencies: list[str],
    *,
    is_optional: bool = False,
) -> int:
    for i, dep in enumerate(dependencies):
        identifier += 1
        if isinstance(dep, str):
            specs = _parse_dependency(
                dep,
                dependencies,
                i,
                "both",
                identifier,
                ignore_pins,
                overwrite_pins_map,
                skip_dependencies,
            )
            for spec in specs:
                _check_allowed_local_dependency(spec.name, is_optional)
                requirements[spec.name].append(spec)
            continue
        assert isinstance(dep, dict)
        for which in ["conda", "pip"]:
            if which in dep:
                specs = _parse_dependency(
                    dep[which],
                    dep,
                    which,
                    which,  # type: ignore[arg-type]
                    identifier,
                    ignore_pins,
                    overwrite_pins_map,
                    skip_dependencies,
                )
                for spec in specs:
                    _check_allowed_local_dependency(spec.name, is_optional)
                    requirements[spec.name].append(spec)
    return identifier


# Alias for backwards compatibility
parse_yaml_requirements = parse_requirements


def _extract_local_dependencies(  # noqa: PLR0912
    path: Path,
    base_path: Path,
    processed: set[Path],
    dependencies: dict[str, set[str]],
    *,
    check_pip_installable: bool = True,
    verbose: bool = False,
    raise_if_missing: bool = True,
    warn_non_managed: bool = True,
) -> None:
    path, extras = parse_folder_or_filename(path)
    if path in processed:
        return
    processed.add(path)
    yaml = YAML(typ="safe")
    data = _load(path, yaml)
    _move_local_optional_dependencies_to_local_dependencies(
        data=data,  # modified in place
        path_with_extras=PathWithExtras(path, extras),
        verbose=verbose,
    )
    # Handle "local_dependencies" (or old name "includes", changed in 0.42.0)
    for local_dependency in _get_local_dependencies(data):
        assert not os.path.isabs(local_dependency)  # noqa: PTH117
        local_path, extras = split_path_and_extras(local_dependency)
        abs_local = (path.parent / local_path).resolve()
        if abs_local.suffix in (".whl", ".zip"):
            if verbose:
                print(f"🔗 Adding `{local_dependency}` from `local_dependencies`")
            dependencies[str(base_path)].add(str(abs_local))
            continue
        if not abs_local.exists():
            if raise_if_missing:
                msg = f"File `{abs_local}` not found."
                raise FileNotFoundError(msg)
            continue

        try:
            requirements_path = parse_folder_or_filename(abs_local).path
        except FileNotFoundError:
            # Means that this is a local package that is not managed by unidep.
            if is_pip_installable(abs_local):
                dependencies[str(base_path)].add(str(abs_local))
                if warn_non_managed:
                    # We do not need to emit this warning when `pip install` is called
                    warn(
                        f"⚠️ Installing a local dependency (`{abs_local.name}`) which"
                        " is not managed by unidep, this will skip all of its"
                        " dependencies, i.e., it will call `pip install` with"
                        "  `--no-deps`. To properly manage this dependency,"
                        " add a `requirements.yaml` or `pyproject.toml` file with"
                        " `[tool.unidep]` in its directory.",
                    )
            elif _is_empty_folder(abs_local):
                msg = (
                    f"`{local_dependency}` in `local_dependencies` is not pip"
                    " installable because it is an empty folder. Is it perhaps"
                    " an uninitialized Git submodule? If so, initialize it with"
                    " `git submodule update --init --recursive`. Otherwise,"
                    " remove it from `local_dependencies`."
                )
                raise RuntimeError(msg) from None
            elif _is_empty_git_submodule(abs_local):
                # Extra check for empty Git submodules (common problem folks run into)
                msg = (
                    f"`{local_dependency}` in `local_dependencies` is not installable"
                    " by pip because it is an empty Git submodule. Either remove it"
                    " from `local_dependencies` or fetch the submodule with"
                    " `git submodule update --init --recursive`."
                )
                raise RuntimeError(msg) from None
            else:
                msg = (
                    f"`{local_dependency}` in `local_dependencies` is not pip"
                    " installable nor is it managed by unidep. Remove it"
                    " from `local_dependencies`."
                )
                raise RuntimeError(msg) from None
            continue

        project_path = str(requirements_path.parent)
        if project_path == str(base_path):
            continue
        if not check_pip_installable or is_pip_installable(requirements_path.parent):
            dependencies[str(base_path)].add(project_path)
        if verbose:
            print(f"🔗 Adding `{requirements_path}` from `local_dependencies`")
        _extract_local_dependencies(
            requirements_path,
            base_path,
            processed,
            dependencies,
            check_pip_installable=check_pip_installable,
            verbose=verbose,
        )


def parse_local_dependencies(
    *paths: Path,
    check_pip_installable: bool = True,
    verbose: bool = False,
    raise_if_missing: bool = True,
    warn_non_managed: bool = True,
) -> dict[Path, list[Path]]:
    """Extract local project dependencies from a list of `requirements.yaml` or `pyproject.toml` files.

    Works by loading the specified `local_dependencies` list.

    Returns a dictionary with the:
    name of the project folder => list of `Path`s of local dependencies folders.
    """  # noqa: E501
    dependencies: dict[str, set[str]] = defaultdict(set)

    for p in paths:
        if verbose:
            print(f"🔗 Analyzing dependencies in `{p}`")
        base_path = p.resolve().parent
        _extract_local_dependencies(
            path=p,
            base_path=base_path,
            processed=set(),
            dependencies=dependencies,
            check_pip_installable=check_pip_installable,
            verbose=verbose,
            raise_if_missing=raise_if_missing,
            warn_non_managed=warn_non_managed,
        )

    return {
        Path(k): sorted({Path(v) for v in v_set})
        for k, v_set in sorted(dependencies.items())
    }


def yaml_to_toml(yaml_path: Path) -> str:
    """Converts a `requirements.yaml` file TOML format."""
    try:
        import tomli_w
    except ImportError:  # pragma: no cover
        msg = (
            "❌ `tomli_w` is required to convert YAML to TOML."
            " Install it with `pip install tomli_w`."
        )
        raise ImportError(msg) from None
    yaml = YAML(typ="rt")
    data = _load(yaml_path, yaml)
    data.pop("name", None)
    dependencies = data.get("dependencies", [])
    for i, dep in enumerate(dependencies):
        if isinstance(dep, str):
            comment = _extract_first_comment(dependencies, i)
            if comment is not None:
                selector = selector_from_comment(comment)
                if selector is not None:
                    dependencies[i] = f"{dep}:{selector}"
            continue
        assert isinstance(dep, dict)
        for which in ["conda", "pip"]:
            if which in dep:
                comment = _extract_first_comment(dep, which)
                if comment is not None:
                    selector = selector_from_comment(comment)
                    if selector is not None:
                        dep[which] = f"{dep[which]}:{selector}"

    return tomli_w.dumps({"tool": {"unidep": data}})


def _is_empty_git_submodule(path: Path) -> bool:
    """Checks if the given path is an empty Git submodule."""
    if not path.is_dir():
        return False

    git_file = path / ".git"
    if not git_file.exists() or not git_file.is_file():
        return False

    # Check if it's empty (apart from the .git file)
    return len(list(path.iterdir())) == 1  # Only .git should be present


def _is_empty_folder(path: Path) -> bool:
    """Checks if the given path is an empty folder."""
    return not any(path.iterdir())



"""Tests for parsing local dependencies from wheels and zips."""

import textwrap
from pathlib import Path
from typing import Literal

import pytest

from unidep import parse_local_dependencies, parse_requirements

from .helpers import maybe_as_toml


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_local_wheel(tmp_path: Path, toml_or_yaml: Literal["toml", "yaml"]) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../example.whl
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)
    r1 = maybe_as_toml(toml_or_yaml, r1)

    local_dep = tmp_path / "example.whl"
    local_dep.touch()  # Create a dummy .whl file

    dependencies = parse_local_dependencies(
        r1,
        check_pip_installable=False,
        verbose=True,
    )
    assert dependencies[project1.resolve()] == [local_dep.resolve()]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_local_zip(tmp_path: Path, toml_or_yaml: Literal["toml", "yaml"]) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../example.zip
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)

    local_dep = tmp_path / "example.zip"
    local_dep.touch()  # Create a dummy .zip file

    dependencies = parse_local_dependencies(r1, check_pip_installable=False)
    assert dependencies[project1.resolve()] == [local_dep.resolve()]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_local_wheel_and_folder(
    tmp_path: Path,
    toml_or_yaml: Literal["toml", "yaml"],
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    project2 = tmp_path / "project2"
    project2.mkdir(exist_ok=True, parents=True)
    (project2 / "setup.py").touch()  # Make project2 pip installable
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../example.whl
                - ../project2
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)

    local_dep = tmp_path / "example.whl"
    local_dep.touch()  # Create a dummy .whl file
    with pytest.warns(UserWarning, match="is not managed by unidep"):
        dependencies = parse_local_dependencies(r1, check_pip_installable=False)
    assert dependencies[project1.resolve()] == [
        local_dep.resolve(),
        project2.resolve(),
    ]

    requirements = parse_requirements(r1, verbose=True)
    assert requirements.requirements == {}


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_local_wheel_with_extras(
    tmp_path: Path,
    toml_or_yaml: Literal["toml", "yaml"],
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../example.whl[extra1,extra2]
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)

    local_dep = tmp_path / "example.whl"
    local_dep.touch()  # Create a dummy .whl file

    dependencies = parse_local_dependencies(r1, check_pip_installable=False)
    assert dependencies[project1.resolve()] == [local_dep.resolve()]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_local_wheel_in_dependencies(
    tmp_path: Path,
    toml_or_yaml: Literal["toml", "yaml"],
) -> None:
    project1 = tmp_path / "project1"
    project1.mkdir(exist_ok=True, parents=True)
    r1 = project1 / "requirements.yaml"
    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../example.whl
            """,
        ),
    )
    r1 = maybe_as_toml(toml_or_yaml, r1)

    local_dep = tmp_path / "example.whl"
    local_dep.touch()  # Create a dummy .whl file

    dependencies = parse_local_dependencies(r1, check_pip_installable=False)
    assert dependencies[project1.resolve()] == [local_dep.resolve()]


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nested_local_dependencies_with_wheel(
    tmp_path: Path,
    toml_or_yaml: Literal["toml", "yaml"],
) -> None:
    project1 = tmp_path / "project1"
    project2 = tmp_path / "project2"
    project3 = tmp_path / "project3"
    for project in [project1, project2, project3]:
        project.mkdir(exist_ok=True, parents=True)
        (project / "setup.py").touch()  # Make projects pip installable

    wheel_dep = tmp_path / "example.whl"
    wheel_dep.touch()  # Create a dummy .whl file

    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"
    r3 = project3 / "requirements.yaml"

    r1.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project2
            """,
        ),
    )

    r2.write_text(
        textwrap.dedent(
            """\
            local_dependencies:
                - ../project3
                - ../example.whl
            """,
        ),
    )

    r3.write_text(
        textwrap.dedent(
            """\
            dependencies:
                - pytest
            """,
        ),
    )

    r1 = maybe_as_toml(toml_or_yaml, r1)
    r2 = maybe_as_toml(toml_or_yaml, r2)
    r3 = maybe_as_toml(toml_or_yaml, r3)

    local_dependencies = parse_local_dependencies(r1, verbose=True)

    assert local_dependencies == {
        project1.resolve(): [
            wheel_dep.resolve(),
            project2.resolve(),
            project3.resolve(),
        ],
    }



"""Test parsing nested local dependencies from YAML files."""

from __future__ import annotations

import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from unidep import (
    parse_local_dependencies,
    parse_requirements,
)

from .helpers import maybe_as_toml

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:  # pragma: no cover
        from typing_extensions import Literal


REPO_ROOT = Path(__file__).parent.parent


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nested_local_dependencies_multiple_levels(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project2 = tmp_path / "project2"
    project3 = tmp_path / "project3"
    project4 = tmp_path / "project4"
    for project in [project1, project2, project3, project4]:
        project.mkdir(exist_ok=True, parents=True)
        (project / "setup.py").touch()  # Make projects pip installable

    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"
    r3 = project3 / "requirements.yaml"
    r4 = project4 / "requirements.yaml"

    r1.write_text(
        textwrap.dedent("""
        dependencies:
            - package1
        local_dependencies:
            - ../project2
    """),
    )

    r2.write_text(
        textwrap.dedent("""
        dependencies:
            - package2
        local_dependencies:
            - ../project3
    """),
    )

    r3.write_text(
        textwrap.dedent("""
        dependencies:
            - package3
        local_dependencies:
            - ../project4
    """),
    )

    r4.write_text(
        textwrap.dedent("""
        dependencies:
            - package4
    """),
    )

    r1 = maybe_as_toml(toml_or_yaml, r1)
    r2 = maybe_as_toml(toml_or_yaml, r2)
    r3 = maybe_as_toml(toml_or_yaml, r3)
    r4 = maybe_as_toml(toml_or_yaml, r4)

    local_dependencies = parse_local_dependencies(
        r1,
        verbose=True,
        check_pip_installable=True,
    )

    assert local_dependencies == {
        project1.resolve(): [
            project2.resolve(),
            project3.resolve(),
            project4.resolve(),
        ],
    }

    requirements = parse_requirements(r1, verbose=True)
    assert set(requirements.requirements.keys()) == {
        "package1",
        "package2",
        "package3",
        "package4",
    }


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nested_local_dependencies_with_circular_reference(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project2 = tmp_path / "project2"
    project3 = tmp_path / "project3"
    for project in [project1, project2, project3]:
        project.mkdir(exist_ok=True, parents=True)
        (project / "setup.py").touch()  # Make projects pip installable

    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"
    r3 = project3 / "requirements.yaml"

    r1.write_text(
        textwrap.dedent("""
        dependencies:
            - package1
        local_dependencies:
            - ../project2
    """),
    )

    r2.write_text(
        textwrap.dedent("""
        dependencies:
            - package2
        local_dependencies:
            - ../project3
    """),
    )

    r3.write_text(
        textwrap.dedent("""
        dependencies:
            - package3
        local_dependencies:
            - ../project1
    """),
    )

    r1 = maybe_as_toml(toml_or_yaml, r1)
    r2 = maybe_as_toml(toml_or_yaml, r2)
    r3 = maybe_as_toml(toml_or_yaml, r3)

    local_dependencies = parse_local_dependencies(
        r1,
        verbose=True,
        check_pip_installable=True,
    )

    assert local_dependencies == {
        project1.resolve(): [project2.resolve(), project3.resolve()],
    }

    requirements = parse_requirements(r1, verbose=True)
    assert set(requirements.requirements.keys()) == {"package1", "package2", "package3"}


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nested_local_dependencies_with_non_unidep_managed_project(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project2 = tmp_path / "project2"
    project3 = tmp_path / "project3"
    for project in [project1, project2]:
        project.mkdir(exist_ok=True, parents=True)
        (project / "setup.py").touch()  # Make projects pip installable

    # Create project3 as a non-unidep managed project
    project3.mkdir(exist_ok=True, parents=True)
    (project3 / "setup.py").touch()  # Make it pip installable but not unidep managed

    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"

    r1.write_text(
        textwrap.dedent("""
        dependencies:
            - package1
        local_dependencies:
            - ../project2
    """),
    )

    r2.write_text(
        textwrap.dedent("""
        dependencies:
            - package2
        local_dependencies:
            - ../project3
    """),
    )

    r1 = maybe_as_toml(toml_or_yaml, r1)
    r2 = maybe_as_toml(toml_or_yaml, r2)

    # project3 is non-unidep managed but pip installable

    with pytest.warns(UserWarning, match="not managed by unidep"):
        local_dependencies = parse_local_dependencies(
            r1,
            verbose=True,
            check_pip_installable=True,
            warn_non_managed=True,
        )

    assert local_dependencies == {
        project1.resolve(): [project2.resolve(), project3.resolve()],
    }

    # We don't expect a warning here anymore, as it should have been raised in parse_local_dependencies
    requirements = parse_requirements(r1, verbose=True)

    assert set(requirements.requirements.keys()) == {"package1", "package2"}


@pytest.mark.parametrize("toml_or_yaml", ["toml", "yaml"])
def test_nested_local_dependencies_with_extras(
    toml_or_yaml: Literal["toml", "yaml"],
    tmp_path: Path,
) -> None:
    project1 = tmp_path / "project1"
    project2 = tmp_path / "project2"
    project3 = tmp_path / "project3"
    for project in [project1, project2, project3]:
        project.mkdir(exist_ok=True, parents=True)
        (project / "setup.py").touch()  # Make projects pip installable

    r1 = project1 / "requirements.yaml"
    r2 = project2 / "requirements.yaml"
    r3 = project3 / "requirements.yaml"

    r1.write_text(
        textwrap.dedent("""
        dependencies:
            - package1
        local_dependencies:
            - ../project2[test,docs]
        optional_dependencies:
            dev:
                - dev-package
    """),
    )

    r2.write_text(
        textwrap.dedent("""
        dependencies:
            - package2
        local_dependencies:
            - ../project3[full]
        optional_dependencies:
            test:
                - pytest
            docs:
                - sphinx
    """),
    )

    r3.write_text(
        textwrap.dedent("""
        dependencies:
            - package3
        optional_dependencies:
            full:
                - extra-package
    """),
    )

    r1 = maybe_as_toml(toml_or_yaml, r1)
    r2 = maybe_as_toml(toml_or_yaml, r2)
    r3 = maybe_as_toml(toml_or_yaml, r3)

    local_dependencies = parse_local_dependencies(
        Path(f"{r1}[dev]"),
        verbose=True,
        check_pip_installable=True,
    )

    assert local_dependencies == {
        project1.resolve(): [project2.resolve(), project3.resolve()],
    }

    requirements = parse_requirements(r1, verbose=True, extras=[["dev"]])
    assert set(requirements.requirements.keys()) == {
        "package1",
        "package2",
        "package3",
        "pytest",
        "sphinx",
        "extra-package",
    }

    # Test with different extras
    requirements_full = parse_requirements(r1, verbose=True, extras=[["dev", "full"]])
    assert set(requirements_full.requirements.keys()) == {
        "package1",
        "package2",
        "package3",
        "pytest",
        "sphinx",
        "extra-package",
    }
    assert requirements_full.optional_dependencies.keys() == {"dev"}
    assert requirements_full.optional_dependencies["dev"].keys() == {"dev-package"}



"""Tests for the `unidep._cli` module (installing conda environment from lock file)."""

import subprocess
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest

from unidep._cli import (
    CondaExecutable,
    _create_env_from_lock,
    _verify_conda_lock_installed,
)


@pytest.fixture
def mock_subprocess_run(monkeypatch: pytest.MonkeyPatch) -> Mock:
    mock = Mock()
    monkeypatch.setattr("subprocess.run", mock)
    return mock


@pytest.fixture
def mock_print(monkeypatch: pytest.MonkeyPatch) -> Mock:
    mock = Mock()
    monkeypatch.setattr("builtins.print", mock)
    return mock


@pytest.mark.parametrize("conda_executable", ["conda", "mamba", "micromamba"])
@pytest.mark.parametrize(
    "env_spec",
    [
        {"conda_env_name": "test_env", "conda_env_prefix": None},
        {"conda_env_name": None, "conda_env_prefix": Path("/path/to/env")},
    ],
)
def test_create_env_from_lock_dry_run(
    conda_executable: CondaExecutable,
    env_spec: dict,
    mock_subprocess_run: Mock,
    mock_print: Mock,
) -> None:
    conda_lock_file = Path("conda-lock.yml")

    with patch("unidep._cli._verify_conda_lock_installed"):
        _create_env_from_lock(
            conda_lock_file=conda_lock_file,
            conda_executable=conda_executable,
            **env_spec,
            dry_run=True,
            verbose=True,
        )

    # Check that subprocess.run was not called
    mock_subprocess_run.assert_not_called()

    # Check that appropriate messages were printed
    env_identifier = (
        f"'{env_spec['conda_env_name']}'"
        if env_spec["conda_env_name"]
        else f"at '{env_spec['conda_env_prefix']}'"
    )

    assert len(mock_print.call_args_list) == 2

    # Check the first message (creating environment)
    first_call = mock_print.call_args_list[0]
    assert first_call.args[0].startswith(
        f"📦 Creating conda environment {env_identifier} with ",
    )

    # Check the command string separately
    cmd_str = first_call.args[0]
    if conda_executable == "micromamba":
        assert "micromamba create" in cmd_str or "micromamba.exe create" in cmd_str
        assert "-f conda-lock.yml" in cmd_str
        assert "--yes" in cmd_str
        assert "--verbose" in cmd_str
    else:
        assert "conda-lock install" in cmd_str
        assert "--log-level=DEBUG" in cmd_str
        if conda_executable == "mamba":
            assert "--mamba" in cmd_str
        elif conda_executable == "conda":
            assert "--conda conda" in cmd_str

    if env_spec["conda_env_name"]:
        assert f"--name {env_spec['conda_env_name']}" in cmd_str
    elif env_spec["conda_env_prefix"]:
        assert f"--prefix {env_spec['conda_env_prefix']}" in cmd_str

    # Check the second message (dry run completed)
    assert mock_print.call_args_list[1] == call(
        "🏁 Dry run completed. No environment was created.",
    )


def test_create_env_from_lock_no_env_specified(mock_print: Mock) -> None:
    conda_lock_file = Path("conda-lock.yml")

    with pytest.raises(SystemExit):
        _create_env_from_lock(
            conda_lock_file=conda_lock_file,
            conda_executable="conda",
            conda_env_name=None,
            conda_env_prefix=None,
            dry_run=True,
            verbose=True,
        )

    mock_print.assert_called_once_with(
        "❌ Please provide either `--conda-env-name` or"
        " `--conda-env-prefix` when using `--conda-lock-file`.",
    )


def test_verify_conda_lock_installed_not_found(
    monkeypatch: pytest.MonkeyPatch,
    mock_print: Mock,
) -> None:
    monkeypatch.setattr("shutil.which", lambda _: None)

    with pytest.raises(SystemExit):
        _verify_conda_lock_installed()

    assert (
        "❌ conda-lock is not installed or not found in PATH."
        in mock_print.call_args[0][0]
    )


def test_verify_conda_lock_installed_not_working(
    monkeypatch: pytest.MonkeyPatch,
    mock_print: Mock,
) -> None:
    monkeypatch.setattr("shutil.which", lambda _: "/path/to/conda-lock")
    monkeypatch.setattr(
        subprocess,
        "run",
        Mock(side_effect=subprocess.CalledProcessError(1, "conda-lock")),
    )

    with pytest.raises(SystemExit):
        _verify_conda_lock_installed()

    assert (
        "❌ conda-lock is installed but not working correctly."
        in mock_print.call_args[0][0]
    )
